import numpy as np
from numpy import array, eye, zeros
from scipy.linalg import block_diag, cholesky


class MerweScaledSigmaPoints:
    """
    Generates sigma points and weights according to Van der Merwe's
    2004 dissertation[1] for the UnscentedKalmanFilter class. It
    parametizes the sigma points using alpha, beta, kappa terms, and
    is the version seen in most publications.
    Unless you know better, this should be your default choice.
    Parameters
    ----------
    n : int
        Dimensionality of the state. 2n+1 weights will be generated.
    alpha : float
        Determins the spread of the sigma points around the mean.
        Usually a small positive value (1e-3) according to [3].
    beta : float
        Incorporates prior knowledge of the distribution of the mean. For
        Gaussian x beta=2 is optimal, according to [3].
    kappa : float, default=0.0
        Secondary scaling parameter usually set to 0 according to [4],
        or to 3-n according to [5].
    sqrt_method : function(ndarray), default=scipy.linalg.cholesky
        Defines how we compute the square root of a matrix, which has
        no unique answer. Cholesky is the default choice due to its
        speed. Typically your alternative choice will be
        scipy.linalg.sqrtm. Different choices affect how the sigma points
        are arranged relative to the eigenvectors of the covariance matrix.
        Usually this will not matter to you; if so the default cholesky()
        yields maximal performance. As of van der Merwe's dissertation of
        2004 [6] this was not a well reseached area so I have no advice
        to give you.
        If your method returns a triangular matrix it must be upper
        triangular. Do not use numpy.linalg.cholesky - for historical
        reasons it returns a lower triangular matrix. The SciPy version
        does the right thing.
    subtract : callable (x, y), optional
        Function that computes the difference between x and y.
        You will have to supply this if your state variable cannot support
        subtraction, such as angles (359-1 degreees is 2, not 358). x and y
        are state vectors, not scalars.
    Attributes
    ----------
    Wm : np.array
        weight for each sigma point for the mean
    Wc : np.array
        weight for each sigma point for the covariance
    Examples
    --------
    See my book Kalman and Bayesian Filters in Python
    https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python
    References
    ----------
    .. [1] R. Van der Merwe "Sigma-Point Kalman Filters for Probabilitic
           Inference in Dynamic State-Space Models" (Doctoral dissertation)
    """

    def __init__(self, n, alpha, beta, kappa, sqrt_method=None, subtract=None):
        # pylint: disable=too-many-arguments

        self.n = n
        self.alpha = alpha
        self.beta = beta
        self.kappa = kappa
        if sqrt_method is None:
            self.sqrt = cholesky
        else:
            self.sqrt = sqrt_method

        if subtract is None:
            self.subtract = np.subtract
        else:
            self.subtract = subtract

        self._compute_weights()

    def num_sigmas(self):
        """ Number of sigma points for each variable in the state x"""
        return 2 * self.n + 1

    def sigma_points(self, x, P):
        """ Computes the sigma points for an unscented Kalman filter
        given the mean (x) and covariance(P) of the filter.
        Returns tuple of the sigma points and weights.
        Works with both scalar and array inputs:
        sigma_points (5, 9, 2) # mean 5, covariance 9
        sigma_points ([5, 2], 9*eye(2), 2) # means 5 and 2, covariance 9I
        Parameters
        ----------
        x : An array-like object of the means of length n
            Can be a scalar if 1D.
            examples: 1, [1,2], np.array([1,2])
        P : scalar, or np.array
           Covariance of the filter. If scalar, is treated as eye(n)*P.
        Returns
        -------
        sigmas : np.array, of size (2n+1, n)
            Two dimensional array of sigma points. Each column contains all of
            the sigmas for one dimension in the problem space.
            Ordered by Xi_0, Xi_{1..n}, Xi_{n+1..2n}
        """

        n = self.n

        if np.isscalar(x):
            x = np.asarray([x])

        if np.isscalar(P):
            P = np.eye(n) * P
        else:
            P = np.atleast_2d(P)

        lambda_ = self.alpha ** 2 * (n + self.kappa) - n
        U = self.sqrt((lambda_ + n) * P)

        sigmas = np.zeros((2 * n + 1, n))
        sigmas[0] = x
        for k in range(n):
            sigmas[k + 1] = self.subtract(x, -U[k])
            sigmas[n + k + 1] = self.subtract(x, U[k])

        return sigmas

    def _compute_weights(self):
        """ Computes the weights for the scaled unscented Kalman filter.
        """

        n = self.n
        lambda_ = self.alpha ** 2 * (n + self.kappa) - n

        c = .5 / (n + lambda_)
        self.Wc = np.full(2 * n + 1, c)
        self.Wm = np.full(2 * n + 1, c)
        self.Wc[0] = lambda_ / (n + lambda_) + (1 - self.alpha ** 2 + self.beta)
        self.Wm[0] = lambda_ / (n + lambda_)


class JulierSigmaPoints(object):
    """
    Generates sigma points and weights according to Simon J. Julier
    and Jeffery K. Uhlmann's original paper[1]. It parametizes the sigma
    points using kappa.
    Parameters
    ----------
    n : int
        Dimensionality of the state. 2n+1 weights will be generated.
    kappa : float, default=0.
        Scaling factor that can reduce high order errors. kappa=0 gives
        the standard unscented filter. According to [Julier], if you set
        kappa to 3-dim_x for a Gaussian x you will minimize the fourth
        order errors in x and P.
    sqrt_method : function(ndarray), default=scipy.linalg.cholesky
        Defines how we compute the square root of a matrix, which has
        no unique answer. Cholesky is the default choice due to its
        speed. Typically your alternative choice will be
        scipy.linalg.sqrtm. Different choices affect how the sigma points
        are arranged relative to the eigenvectors of the covariance matrix.
        Usually this will not matter to you; if so the default cholesky()
        yields maximal performance. As of van der Merwe's dissertation of
        2004 [6] this was not a well reseached area so I have no advice
        to give you.
        If your method returns a triangular matrix it must be upper
        triangular. Do not use numpy.linalg.cholesky - for historical
        reasons it returns a lower triangular matrix. The SciPy version
        does the right thing.
    subtract : callable (x, y), optional
        Function that computes the difference between x and y.
        You will have to supply this if your state variable cannot support
        subtraction, such as angles (359-1 degreees is 2, not 358). x and y
    Attributes
    ----------
    Wm : np.array
        weight for each sigma point for the mean
    Wc : np.array
        weight for each sigma point for the covariance
    References
    ----------
    .. [1] Julier, Simon J.; Uhlmann, Jeffrey "A New Extension of the Kalman
        Filter to Nonlinear Systems". Proc. SPIE 3068, Signal Processing,
        Sensor Fusion, and Target Recognition VI, 182 (July 28, 1997)
   """

    def __init__(self, n, kappa=0., sqrt_method=None, subtract=None):

        self.n = n
        self.kappa = kappa
        if sqrt_method is None:
            self.sqrt = cholesky
        else:
            self.sqrt = sqrt_method

        if subtract is None:
            self.subtract = np.subtract
        else:
            self.subtract = subtract

        self._compute_weights()

    def num_sigmas(self):
        """ Number of sigma points for each variable in the state x"""
        return 2 * self.n + 1

    def sigma_points(self, x, P):
        r""" Computes the sigma points for an unscented Kalman filter
        given the mean (x) and covariance(P) of the filter.
        kappa is an arbitrary constant. Returns sigma points.
        Works with both scalar and array inputs:
        sigma_points (5, 9, 2) # mean 5, covariance 9
        sigma_points ([5, 2], 9*eye(2), 2) # means 5 and 2, covariance 9I
        Parameters
        ----------
        x : array-like object of the means of length n
            Can be a scalar if 1D.
            examples: 1, [1,2], np.array([1,2])
        P : scalar, or np.array
           Covariance of the filter. If scalar, is treated as eye(n)*P.
        kappa : float
            Scaling factor.
        Returns
        -------
        sigmas : np.array, of size (2n+1, n)
            2D array of sigma points :math:`\chi`. Each column contains all of
            the sigmas for one dimension in the problem space. They
            are ordered as:
            .. math::
                :nowrap:
                \begin{eqnarray}
                  \chi[0]    = &x \\
                  \chi[1..n] = &x + [\sqrt{(n+\kappa)P}]_k \\
                  \chi[n+1..2n] = &x - [\sqrt{(n+\kappa)P}]_k
                \end{eqnarray}
        """

        if self.n != np.size(x):
            raise ValueError("expected size(x) {}, but size is {}".format(
                self.n, np.size(x)))

        n = self.n

        if np.isscalar(x):
            x = np.asarray([x])

        n = np.size(x)  # dimension of problem

        if np.isscalar(P):
            P = np.eye(n) * P
        else:
            P = np.atleast_2d(P)

        sigmas = np.zeros((2 * n + 1, n))

        # implements U'*U = (n+kappa)*P. Returns lower triangular matrix.
        # Take transpose so we can access with U[i]
        U = self.sqrt((n + self.kappa) * P)

        sigmas[0] = x
        for k in range(n):
            # pylint: disable=bad-whitespace
            sigmas[k + 1] = self.subtract(x, -U[k])
            sigmas[n + k + 1] = self.subtract(x, U[k])
        return sigmas

    def _compute_weights(self):
        """ Computes the weights for the unscented Kalman filter. In this
        formulation the weights for the mean and covariance are the same.
        """

        n = self.n
        k = self.kappa

        self.Wm = np.full(2 * n + 1, .5 / (n + k))
        self.Wm[0] = k / (n + k)
        self.Wc = self.Wm


class SimplexSigmaPoints(object):
    """
    Generates sigma points and weights according to the simplex
    method presented in [1].
    Parameters
    ----------
    n : int
        Dimensionality of the state. n+1 weights will be generated.
    sqrt_method : function(ndarray), default=scipy.linalg.cholesky
        Defines how we compute the square root of a matrix, which has
        no unique answer. Cholesky is the default choice due to its
        speed. Typically your alternative choice will be
        scipy.linalg.sqrtm
        If your method returns a triangular matrix it must be upper
        triangular. Do not use numpy.linalg.cholesky - for historical
        reasons it returns a lower triangular matrix. The SciPy version
        does the right thing.
    subtract : callable (x, y), optional
        Function that computes the difference between x and y.
        You will have to supply this if your state variable cannot support
        subtraction, such as angles (359-1 degreees is 2, not 358). x and y
        are state vectors, not scalars.
    Attributes
    ----------
    Wm : np.array
        weight for each sigma point for the mean
    Wc : np.array
        weight for each sigma point for the covariance
    References
    ----------
    .. [1] Phillippe Moireau and Dominique Chapelle "Reduced-Order
           Unscented Kalman Filtering with Application to Parameter
           Identification in Large-Dimensional Systems"
           DOI: 10.1051/cocv/2010006
    """

    def __init__(self, n, alpha=1, sqrt_method=None, subtract=None):
        self.n = n
        self.alpha = alpha
        if sqrt_method is None:
            self.sqrt = cholesky
        else:
            self.sqrt = sqrt_method

        if subtract is None:
            self.subtract = np.subtract
        else:
            self.subtract = subtract

        self._compute_weights()

    def num_sigmas(self):
        """ Number of sigma points for each variable in the state x"""
        return self.n + 1

    def sigma_points(self, x, P):
        """
        Computes the implex sigma points for an unscented Kalman filter
        given the mean (x) and covariance(P) of the filter.
        Returns tuple of the sigma points and weights.
        Works with both scalar and array inputs:
        sigma_points (5, 9, 2) # mean 5, covariance 9
        sigma_points ([5, 2], 9*eye(2), 2) # means 5 and 2, covariance 9I
        Parameters
        ----------
        x : An array-like object of the means of length n
            Can be a scalar if 1D.
            examples: 1, [1,2], np.array([1,2])
        P : scalar, or np.array
           Covariance of the filter. If scalar, is treated as eye(n)*P.
        Returns
        -------
        sigmas : np.array, of size (n+1, n)
            Two dimensional array of sigma points. Each column contains all of
            the sigmas for one dimension in the problem space.
            Ordered by Xi_0, Xi_{1..n}
        """

        if self.n != np.size(x):
            raise ValueError("expected size(x) {}, but size is {}".format(
                self.n, np.size(x)))

        n = self.n

        if np.isscalar(x):
            x = np.asarray([x])
        x = x.reshape(-1, 1)

        if np.isscalar(P):
            P = np.eye(n) * P
        else:
            P = np.atleast_2d(P)

        U = self.sqrt(P)

        lambda_ = n / (n + 1)
        Istar = np.array([[-1 / np.sqrt(2 * lambda_), 1 / np.sqrt(2 * lambda_)]])

        for d in range(2, n + 1):
            row = np.ones((1, Istar.shape[1] + 1)) * 1. / np.sqrt(
                lambda_ * d * (d + 1))  # pylint: disable=unsubscriptable-object
            row[0, -1] = -d / np.sqrt(lambda_ * d * (d + 1))
            Istar = np.r_[np.c_[Istar, np.zeros((Istar.shape[0]))], row]  # pylint: disable=unsubscriptable-object

        I = np.sqrt(n) * Istar
        scaled_unitary = (U.T).dot(I)

        sigmas = self.subtract(x, -scaled_unitary)
        return sigmas.T

    def _compute_weights(self):
        """ Computes the weights for the scaled unscented Kalman filter. """

        n = self.n
        c = 1. / (n + 1)
        self.Wm = np.full(n + 1, c)
        self.Wc = self.Wm


def order_by_derivative(Q, dim, block_size):
    """
    Given a matrix Q, ordered assuming state space
        [x y z x' y' z' x'' y'' z''...]
    return a reordered matrix assuming an ordering of
       [ x x' x'' y y' y'' z z' y'']
    This works for any covariance matrix or state transition function
    Parameters
    ----------
    Q : np.array, square
        The matrix to reorder
    dim : int >= 1
       number of independent state variables. 3 for x, y, z
    block_size : int >= 0
        Size of derivatives. Second derivative would be a block size of 3
        (x, x', x'')
    """

    N = dim * block_size

    D = zeros((N, N))

    Q = array(Q)
    for i, x in enumerate(Q.ravel()):
        f = eye(block_size) * x

        ix, iy = (i // dim) * block_size, (i % dim) * block_size
        D[ix:ix + block_size, iy:iy + block_size] = f

    return D


def Q_discrete_white_noise(dim, dt=1., var=1., block_size=1, order_by_dim=True):
    """
    Returns the Q matrix for the Discrete Constant White Noise
    Model. dim may be either 2, 3, or 4 dt is the time step, and sigma
    is the variance in the noise.
    Q is computed as the G * G^T * variance, where G is the process noise per
    time step. In other words, G = [[.5dt^2][dt]]^T for the constant velocity
    model.
    Parameters
    -----------
    dim : int (2, 3, or 4)
        dimension for Q, where the final dimension is (dim x dim)
    dt : float, default=1.0
        time step in whatever units your filter is using for time. i.e. the
        amount of time between innovations
    var : float, default=1.0
        variance in the noise
    block_size : int >= 1
        If your state variable contains more than one dimension, such as
        a 3d constant velocity model [x x' y y' z z']^T, then Q must be
        a block diagonal matrix.
    order_by_dim : bool, default=True
        Defines ordering of variables in the state vector. `True` orders
        by keeping all derivatives of each dimensions)
        [x x' x'' y y' y'']
        whereas `False` interleaves the dimensions
        [x y z x' y' z' x'' y'' z'']

    """

    if dim not in [2, 3, 4]:
        raise ValueError("dim must be between 2 and 4")

    if dim == 2:
        Q = [[.25 * dt ** 4, .5 * dt ** 3],
             [.5 * dt ** 3, dt ** 2]]
    elif dim == 3:
        Q = [[.25 * dt ** 4, .5 * dt ** 3, .5 * dt ** 2],
             [.5 * dt ** 3, dt ** 2, dt],
             [.5 * dt ** 2, dt, 1]]
    else:
        Q = [[(dt ** 6) / 36, (dt ** 5) / 12, (dt ** 4) / 6, (dt ** 3) / 6],
             [(dt ** 5) / 12, (dt ** 4) / 4, (dt ** 3) / 2, (dt ** 2) / 2],
             [(dt ** 4) / 6, (dt ** 3) / 2, dt ** 2, dt],
             [(dt ** 3) / 6, (dt ** 2) / 2, dt, 1.]]

    if order_by_dim:
        return block_diag(*[Q] * block_size) * var
    return order_by_derivative(array(Q), dim, block_size) * var



import numpy as np
import pandas as pd
from sklearn.feature_selection import VarianceThreshold

from fedot_ind.core.operation.transformation.WindowSelection import WindowSizeSelection


class FeatureSpaceReductor:

    def reduce_feature_space(self, features: pd.DataFrame,
                             var_threshold: float = 0.01,
                             corr_threshold: float = 0.98) -> pd.DataFrame:
        """Method responsible for reducing feature space.

        Args:
            features: dataframe with extracted features.
            corr_threshold: cut-off value for correlation threshold.
            var_threshold: cut-off value for variance threshold.

        Returns:
            Dataframe with reduced feature space.

        """
        init_feature_space_size = features.shape[1]

        features = self._drop_stable_features(features, var_threshold)
        features_new = self._drop_correlated_features(corr_threshold, features)

        final_feature_space_size = features_new.shape[1]

        if init_feature_space_size != final_feature_space_size:
            self.logger.info(f'Feature space reduced from {init_feature_space_size} to {final_feature_space_size}')

        return features_new

    def _drop_correlated_features(self, corr_threshold, features):
        features_corr = features.corr(method='pearson')
        mask = np.ones(features_corr.columns.size) - np.eye(features_corr.columns.size)
        df_corr = mask * features_corr
        drops = []
        for col in df_corr.columns.values:
            # continue if the feature is already in the drop list
            if np.in1d([col], drops):
                continue

            index_of_corr_feature = df_corr[abs(df_corr[col]) > corr_threshold].index
            drops = np.union1d(drops, index_of_corr_feature)

        if len(drops) == 0:
            self.logger.info('No correlated features found')
            return features

        features_new = features.copy()
        features_new.drop(drops, axis=1, inplace=True)
        return features_new

    def _drop_stable_features(self, features, var_threshold):
        try:
            variance_reducer = VarianceThreshold(threshold=var_threshold)
            variance_reducer.fit_transform(features)
            unstable_features_mask = variance_reducer.get_support()
            features = features.loc[:, unstable_features_mask]
        except ValueError:
            self.logger.info('Variance reducer has not found any features with low variance')
        return features

    def validate_window_size(self, ts: np.ndarray):
        if self.window_size is None or self.window_size > ts.shape[0] / 2:
            self.logger.info('Window size is not defined or too big (> ts_length/2)')
            self.window_size, _ = WindowSizeSelection(time_series=ts).get_window_size()
            self.logger.info(f'Window size was set to {self.window_size}')



import pathlib
import random
from typing import Sequence

from fedot.core.composer.gp_composer.specific_operators import parameter_change_mutation, boosting_mutation
from fedot.core.pipelines.node import PipelineNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.search_space import SearchSpace
from fedot.core.repository.operation_types_repository import OperationTypesRepository, get_operations_for_task
from fedot.core.repository.tasks import Task, TaskTypesEnum
from golem.core.dag.verification_rules import ERROR_PREFIX
from golem.core.optimisers.genetic.operators.mutation import MutationTypesEnum

from fedot_ind.core.architecture.utils.utils import PROJECT_PATH
from fedot_ind.core.tuning.search_space import get_industrial_search_space


def add_preprocessing(pipeline: Pipeline, **kwargs) -> Pipeline:
    task = Task(TaskTypesEnum.classification)
    basis_models = get_operations_for_task(task=task, mode='data_operation', tags=["basis"])
    extractors = get_operations_for_task(task=task, mode='data_operation', tags=["extractor"])
    models = get_operations_for_task(task=task, mode='model')
    basis_model = PipelineNode(random.choice(basis_models))
    extractor_model = PipelineNode(random.choice(extractors), nodes_from=[basis_model])

    try:
        node_to_mutate = list(filter(lambda x: x.name  in models, pipeline.nodes))[0]
    except:
        pipeline.show()
    if node_to_mutate.nodes_from:
        node_to_mutate.nodes_from.append(extractor_model)
    else:
        node_to_mutate.nodes_from = [extractor_model]
    pipeline.nodes.append(basis_model)
    pipeline.nodes.append(extractor_model)

    return pipeline


def _get_default_industrial_mutations(task_type: TaskTypesEnum) -> Sequence[MutationTypesEnum]:
    mutations = [parameter_change_mutation,
                 MutationTypesEnum.single_change,
                 add_preprocessing
                 ]
    return mutations


def _get_default_mutations(task_type: TaskTypesEnum) -> Sequence[MutationTypesEnum]:
    mutations = [parameter_change_mutation,
                 MutationTypesEnum.single_change,
                 MutationTypesEnum.single_drop,
                 MutationTypesEnum.single_add]

    # TODO remove workaround after boosting mutation fix
    if task_type == TaskTypesEnum.ts_forecasting:
        mutations.append(boosting_mutation)
    # TODO remove workaround after validation fix
    if task_type is not TaskTypesEnum.ts_forecasting:
        mutations.append(MutationTypesEnum.single_edge)

    return mutations


def has_no_data_flow_conflicts_in_industrial_pipeline(pipeline: Pipeline):
    """ Function checks the correctness of connection between nodes """
    task = Task(TaskTypesEnum.classification)
    basis_models = get_operations_for_task(task=task, mode='data_operation', tags=["basis"])
    extractor = get_operations_for_task(task=task, mode='data_operation', tags=["extractor"])
    other = get_operations_for_task(task=task, forbidden_tags=["basis", "extractor"])

    for node in pipeline.nodes:
        # Operation name in the current node
        current_operation = node.operation.operation_type
        parent_nodes = node.nodes_from
        if parent_nodes:
            if current_operation in basis_models:
                raise ValueError(
                    f'{ERROR_PREFIX} Pipeline has incorrect subgraph with wrong parent nodes combination')
            # There are several parents for current node or at least 1
            for parent in parent_nodes:
                parent_operation = parent.operation.operation_type
                if current_operation in extractor:
                    if parent_operation not in basis_models:
                        raise ValueError(
                            f'{ERROR_PREFIX} Pipeline has incorrect subgraph with wrong parent nodes combination')
                elif current_operation in other:
                    if parent_operation in basis_models:
                        raise ValueError(
                            f'{ERROR_PREFIX} Pipeline has incorrect subgraph with wrong parent nodes combination')
        else:
            # Only basis models can be primary
            if current_operation not in basis_models:
                raise ValueError(
                    f'{ERROR_PREFIX} Pipeline has incorrect subgraph with wrong parent nodes combination')
    return True


class IndustrialModels:
    def __init__(self):
        self.industrial_data_operation_path = pathlib.Path(PROJECT_PATH,
                                                           'core',
                                                           'repository',
                                                           'data',
                                                           'industrial_data_operation_repository.json')
        self.base_data_operation_path = pathlib.Path('data_operation_repository.json')

    def __enter__(self):
        """
        Switching to industrial models
        """
        OperationTypesRepository.__repository_dict__.update({'data_operation':
                                                                 {'file': self.industrial_data_operation_path,
                                                                  'initialized_repo': None,
                                                                  'default_tags': []}})
        OperationTypesRepository.assign_repo('data_operation', self.industrial_data_operation_path)
        setattr(SearchSpace, "get_parameters_dict", get_industrial_search_space)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Switching to fedot models.
        """
        OperationTypesRepository.__repository_dict__.update({'data_operation':
                                                                 {'file': self.base_data_operation_path,
                                                                  'initialized_repo': None,
                                                                  'default_tags': [
                                                                      OperationTypesRepository.DEFAULT_DATA_OPERATION_TAGS]}})
        OperationTypesRepository.assign_repo('data_operation', self.base_data_operation_path)

        #setattr(ApiComposer, "_get_default_mutations", _get_default_mutations)



import logging
import os

import pandas as pd

from fedot_ind.core.architecture.utils.utils import default_path_to_save_results


class ResultSaver:
    def __init__(self, dataset_name: str, generator_name: str, output_dir: str = None):
        if generator_name is None:
            generator_name = 'without_generator'
        self.path = self.__init_save_path(dataset_name, generator_name, output_dir)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.save_method_dict = {'labels': self.save_labels,
                                 'probs': self.save_probs,
                                 'metrics': self.save_metrics}

    def __init_save_path(self, dataset_name, generator_name, output_dir):
        if output_dir is None:
            self.output_dir = default_path_to_save_results()
        else:
            self.output_dir = os.path.abspath(output_dir)
        path = os.path.join(self.output_dir, generator_name, dataset_name)
        os.makedirs(path, exist_ok=True)

        return path

    def save(self, predicted_data, prediction_type: str):
        self.logger.info(f'Saving predicted {prediction_type} to {self.path}')
        try:
            self.save_method_dict[prediction_type](predicted_data)
        except Exception:
            self.logger.error(f'Can not save {prediction_type} type to {self.path}')

    def save_labels(self, label_data):
        df = pd.DataFrame(label_data, dtype=int)
        df.to_csv(os.path.join(self.path, 'predicted_labels.csv'))

    def save_probs(self, prob_data):
        df_preds = pd.DataFrame(prob_data.round(3), dtype=float)
        df_preds.columns = [f'Class_{x + 1}' for x in df_preds.columns]
        df_preds.to_csv(os.path.join(self.path, 'predicted_probs.csv'))

    def save_metrics(self, metrics: dict):
        df = pd.DataFrame(metrics, index=[0])
        df.to_csv(os.path.join(self.path, 'metrics.csv'))




from abc import ABC
import pandas as pd

class FeatureBuilderSelector:
    """Selects the appropriate feature builder based on the operation name.

    Args:
        operation_name (str): Name of the operation.
        feature_generator (callable): Function that generates features.

    """
    def __init__(self, operation_name: str, feature_generator: callable):
        self.operation_name = operation_name
        self.feature_generator = feature_generator

    def select_transformation(self):
        if self.operation_name.startswith('window'):
            return WindowBuilder(self.feature_generator).add_transformation
        elif self.operation_name.startswith('random'):
            return RandomBuilder(self.feature_generator).add_transformation
        else:
            return SteadyBuilder(self.feature_generator).add_transformation


class BuilderBase(ABC):
    """Abstract class for feature builders.

    """
    def __init__(self, feature_generator: callable):
        self.feature_generator = feature_generator

    def add_transformation(self, feature_array: pd.DataFrame, window_size: None):
        pass


class WindowBuilder(BuilderBase):
    def add_transformation(self, feature_array: pd.DataFrame, window_size: None):
        if window_size is None:
            self.window_size = round(feature_array.shape[1] / 10)
        subseq_generator = range(0, feature_array.shape[1], self.window_size)
        slice_ts = [feature_array.iloc[:, i:i + self.window_size] for i in subseq_generator]
        slice_ts = list(filter(lambda x: x.shape[1] > 1, slice_ts))
        feature_list = map(lambda x: self.feature_generator(x), slice_ts)
        X = pd.concat(feature_list)
        del feature_list
        return X


class RandomBuilder(BuilderBase):
    def add_transformation(self, feature_generator: callable, **kwargs):
        pass


class SteadyBuilder(BuilderBase):
    def add_transformation(self, feature_array: pd.DataFrame, window_size: None):
        X = pd.concat(self.feature_generator(feature_array))
        return X



from pymonad.list import ListMonad
from pymonad.either import Right
from fedot_ind.core.architecture.pipelines.abstract_pipeline import AbstractPipelines
from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from fedot_ind.core.operation.transformation.basis.data_driven import DataDrivenBasisImplementation
from functools import partial


class ClassificationPipelines(AbstractPipelines):

    def __call__(self, pipeline_type: str = 'SpecifiedFeatureGeneratorTSC'):
        pipeline_dict = {'DataDrivenTSC': self.__ts_data_driven_pipeline,
                         'DataDrivenMultiTSC': self.__multits_data_driven_pipeline,
                         'SpecifiedBasisTSC': self.__specified_basis_pipeline,
                         'SpecifiedFeatureGeneratorTSC': self.__specified_fg_pipeline,
                         }
        return pipeline_dict[pipeline_type]

    def get_feature_generator(self, **kwargs):
        lambda_func_dict = kwargs['steps']
        lambda_func_dict['concatenate_features'] = lambda x: list_of_features.append(x)
        input_data = kwargs['input_data']
        list_of_features = []

        if kwargs['pipeline_type'] == 'DataDrivenTSC':
            pipeline = Right(input_data).then(lambda_func_dict['create_list_of_ts']).map(
                lambda_func_dict['transform_to_basis']).map(lambda_func_dict['reduce_basis']).map(
                lambda_func_dict['extract_features']).then(lambda_func_dict['concatenate_features']).insert(
                self._get_feature_matrix(list_of_features=list_of_features, mode='1D'))

        elif kwargs['pipeline_type'] == 'SpecifiedFeatureGeneratorTSC':
            pipeline = Right(input_data).then(lambda_func_dict['create_list_of_ts']).\
                map(lambda_func_dict['extract_features']).\
                then(lambda_func_dict['concatenate_features']).\
                insert(self._get_feature_matrix(list_of_features=list_of_features, mode='1D'))

        elif kwargs['pipeline_type'] == 'DataDrivenMultiTSC':
            pipeline = Right(input_data).then(lambda_func_dict['create_list_of_ts']).map(
                lambda_func_dict['transform_to_basis']).then(lambda_func_dict['reduce_basis']).then(
                lambda_func_dict['extract_features']).then(lambda_func_dict['concatenate_features']).insert(
                self._get_feature_matrix(list_of_features=list_of_features, mode=kwargs['ensemble']))

        return pipeline

    def __ts_data_driven_pipeline(self, **kwargs):
        feature_extractor, classificator, lambda_func_dict = self._init_pipeline_nodes(**kwargs)
        data_basis = DataDrivenBasisImplementation(kwargs['data_driven_hyperparams'])
        n_components = kwargs['data_driven_hyperparams']['n_components']
        lambda_func_dict['transform_to_basis'] = lambda \
                x: self.basis if self.basis is not None else data_basis._transform(x)
        lambda_func_dict['reduce_basis'] = lambda x: x[:data_basis.min_rank, :] if n_components \
                                                                                   is None else x[:n_components, :]
        train_feature_generator_module = self.get_feature_generator(input_data=self.train_features,
                                                                    steps=lambda_func_dict,
                                                                    pipeline_type='DataDrivenTSC')
        classificator_module = train_feature_generator_module.then(lambda_func_dict['fit_model'])

        test_feature_generator_module = self.get_feature_generator(input_data=self.test_features,
                                                                   steps=lambda_func_dict,
                                                                   pipeline_type='DataDrivenTSC')
        prediction = test_feature_generator_module.then(lambda_func_dict['predict'])
        classificator.feature_generator = partial(self.get_feature_generator, steps=lambda_func_dict,
                                                  pipeline_type='DataDrivenTSC')
        self.basis = data_basis.basis
        return classificator, prediction.value[0]

    def __multits_data_driven_pipeline(self, ensemble: str = 'Multi', **kwargs):
        feature_extractor, classificator, lambda_func_dict = self._init_pipeline_nodes(**kwargs)
        data_basis = DataDrivenBasisImplementation(kwargs['data_driven_hyperparams'])

        lambda_func_dict['transform_to_basis'] = lambda x: self.basis if self.basis is not None else data_basis.fit(x)
        lambda_func_dict['reduce_basis'] = lambda list_of_components: ListMonad(
            [component[:, 0] if component.shape[1] == 1 else
             component[:, rank]
             for component, rank in zip(list_of_components, kwargs['component'])])
        lambda_func_dict['extract_features'] = lambda list_of_components: ListMonad(
            [feature_extractor.get_features(component)
             for component in list_of_components])

        if ensemble == 'MultiEnsemble':
            lambda_func_dict['fit_model'] = lambda feature_by_each_component: ListMonad({str(index):
                classificator.fit(
                    train_features=feature_set, train_target=self.train_target) for index, feature_set in
                enumerate(feature_by_each_component)})
            lambda_func_dict['predict'] = lambda feature_by_each_component: ListMonad(
                {index: {'predicted_labels': classificator.predict(features=feature_set),
                         'predicted_probs': classificator.predict_proba(
                             features=feature_set)} for (index, classificator), feature_set in
                 zip(classificator_module.value[0].items(), feature_by_each_component)})

        train_feature_generator_module = self.get_feature_generator(input_data=self.train_features,
                                                                    steps=lambda_func_dict,
                                                                    pipeline_type='DataDrivenMultiTSC',
                                                                    ensemble=ensemble)

        classificator_module = train_feature_generator_module.then(lambda_func_dict['fit_model'])

        test_feature_generator_module = self.get_feature_generator(input_data=self.test_features,
                                                                   steps=lambda_func_dict,
                                                                   pipeline_type='DataDrivenMultiTSC',
                                                                   ensemble=ensemble)
        prediction = test_feature_generator_module.then(lambda_func_dict['predict'])
        classificator.feature_generator = partial(self.get_feature_generator, steps=lambda_func_dict,
                                                  pipeline_type='DataDrivenMultiTSC')
        self.basis = data_basis.basis

        return classificator, prediction.value[0]

    def __specified_basis_pipeline(self, **kwargs):
        pass

    def __specified_fg_pipeline(self, **kwargs):
        feature_extractor, classificator, lambda_func_dict = self._init_pipeline_nodes(**kwargs)

        train_feature_generator_module = self.get_feature_generator(input_data=self.train_features,
                                                                    steps=lambda_func_dict,
                                                                    pipeline_type='SpecifiedFeatureGeneratorTSC')
        classificator_module = train_feature_generator_module.then(lambda_func_dict['fit_model'])

        test_feature_generator_module = self.get_feature_generator(input_data=self.test_features,
                                                                   steps=lambda_func_dict,
                                                                   pipeline_type='SpecifiedFeatureGeneratorTSC')
        prediction = test_feature_generator_module.then(lambda_func_dict['predict'])
        classificator.feature_generator = partial(self.get_feature_generator, steps=lambda_func_dict,
                                                  pipeline_type='SpecifiedFeatureGeneratorTSC')
        return classificator, prediction.value[0]


if __name__ == "__main__":
    train_data, test_data = DataLoader(dataset_name='ECG200').load_data()
    model_hyperparams = {
        'problem': 'classification',
        'seed': 42,
        'timeout': 1,
        'max_depth': 10,
        'max_arity': 4,
        'cv_folds': 2,
        'logging_level': 20,
        'n_jobs': 2
    }

    pipeline = ClassificationPipelines(train_data=train_data, test_data=test_data).__call__('DataDrivenMultiTSC')
    pipeline(feature_generator_type='statistical',
             model_hyperparams=model_hyperparams,
             feature_hyperparams=None,
             data_driven_hyperparams={'n_components': 3, 'window_size': 30})



import pandas as pd

from fedot_ind.core.models.BaseExtractor import BaseExtractor


class WindowedFeatureExtractor(BaseExtractor):
    @staticmethod
    def apply_window_for_stat_feature(ts_data: pd.DataFrame,
                                      feature_generator: callable,
                                      window_size: int = None):
        ts_data = ts_data.T
        if window_size is None:
            window_size = round(ts_data.shape[1] / 10)
        else:
            window_size = round(ts_data.shape[1] / window_size)
        tmp_list = []
        for i in range(0, ts_data.shape[1], window_size):
            slice_ts = ts_data.iloc[:, i:i + window_size]
            if slice_ts.shape[1] == 1:
                break
            else:
                df = feature_generator(slice_ts)
                df.columns = [x + f'_on_interval: {i} - {i + window_size}' for x in df.columns]
                tmp_list.append(df)
        return tmp_list


import logging
import os
import shutil
import urllib.request as request
import zipfile

import numpy as np
import pandas as pd
from scipy.io.arff import loadarff
from sktime.datasets._data_io import load_from_tsfile_to_dataframe

from fedot_ind.core.architecture.utils.utils import PROJECT_PATH


class DataLoader:
    """Class for reading data from ``tsv`` files and downloading from UCR archive if not found locally.
    At the moment supports only ``.txt`` and ``.arff`` formats, but not relational ``.arff`` or ``.ts`` files.

    Args:
        dataset_name: name of dataset

    Examples:
        >>> data_loader = DataLoader('ItalyPowerDemand')
        >>> train_data, test_data = data_loader.load_data()
    """

    def __init__(self, dataset_name: str):
        self.logger = logging.getLogger('DataLoader')
        self.dataset_name = dataset_name

    def load_data(self) -> tuple:
        """Load data for classification experiment locally or externally from UCR archive.

        Returns:
            tuple: train and test data
        """
        dataset_name = self.dataset_name
        self.logger.info(f'Trying to read {dataset_name} data locally')
        train_data, test_data = self.read_tsv(dataset_name)
        if train_data is None:
            train_data, test_data = self.read_tstime(dataset_name)

        if train_data is None:
            self.logger.info(f'Dataset {dataset_name} not found in data folder. Downloading...')

            # Create temporary folder for downloaded data
            cache_path = os.path.join(PROJECT_PATH, 'temp_cache/')
            download_path = cache_path + 'downloads/'
            temp_data_path = cache_path + 'temp_data/'
            filename = 'temp_data_{}'.format(dataset_name)
            for _ in (download_path, temp_data_path):
                os.makedirs(_, exist_ok=True)

            url = f"http://www.timeseriesclassification.com/Downloads/{dataset_name}.zip"
            request.urlretrieve(url, download_path + filename)
            try:
                zipfile.ZipFile(download_path + filename).extractall(temp_data_path)
            except zipfile.BadZipFile:
                self.logger.error(f'Cannot extract data: {dataset_name} dataset not found in UCR archive')
                return None, None

            self.logger.info(f'{dataset_name} data downloaded. Unpacking...')
            train_data, test_data = self.extract_data(dataset_name, temp_data_path)

            shutil.rmtree(cache_path)
            return train_data, test_data

        return train_data, test_data

    def extract_data(self, dataset_name: str, temp_data_path: str):
        """Unpacks data from downloaded file and saves it into Data folder with ``.tsv`` extension.

        Args:
            dataset_name: name of dataset
            temp_data_path: path to temporary folder with downloaded data

        Returns:
            tuple: train and test data

        """

        # If data unpacked as .txt file
        try:
            if os.path.isfile(temp_data_path + '/' + dataset_name + '_TRAIN.txt'):
                x_test, x_train, y_test, y_train = self.read_txt(dataset_name, temp_data_path)

            # If data unpacked as .arff file
            elif os.path.isfile(temp_data_path + '/' + dataset_name + '_TRAIN.arff'):
                x_test, x_train, y_test, y_train = self.read_arff(dataset_name, temp_data_path)

            elif os.path.isfile(temp_data_path + '/' + dataset_name + '_TRAIN.ts'):

                x_test, y_test = load_from_tsfile_to_dataframe(temp_data_path + '/' + dataset_name + '_TEST.ts',
                                                               return_separate_X_and_y=True)
                x_train, y_train = load_from_tsfile_to_dataframe(temp_data_path + '/' + dataset_name + '_TRAIN.ts',
                                                                 return_separate_X_and_y=True)
            else:
                self.logger.error('Loaded data is of unknown format')
                return None, None
        except Exception as e:
            self.logger.error(f'Error while unpacking data: {e}')
            return None, None

        # Conversion of target values to int or str
        try:
            y_train = y_train.astype('int64')
            y_test = y_test.astype('int64')
        except ValueError:
            y_train = y_train.astype(str)
            y_test = y_test.astype(str)

        # Save data to tsv files
        data_path = os.path.join(PROJECT_PATH, 'data', dataset_name)
        os.makedirs(data_path, exist_ok=True)

        self.logger.info(f'Saving {dataset_name} data to tsv files')
        for subset in ('TRAIN', 'TEST'):
            df = pd.DataFrame(x_train if subset == 'TRAIN' else x_test)
            df.insert(0, 'class', y_train if subset == 'TRAIN' else y_test)
            df.to_csv(os.path.join(data_path, f'{dataset_name}_{subset}.tsv'), sep='\t', index=False, header=False)

        del df
        return (pd.DataFrame(x_train), y_train), (pd.DataFrame(x_test), y_test)

    def read_arff(self, dataset_name, temp_data_path):
        """Reads data from ``.arff`` file.

        """
        train = loadarff(temp_data_path + dataset_name + '_TRAIN.arff')
        test = loadarff(temp_data_path + dataset_name + '_TEST.arff')
        try:
            data_train = np.asarray([train[0][name] for name in train[1].names()])
            x_train = data_train[:-1].T.astype('float64')
            y_train = data_train[-1]

            data_test = np.asarray([test[0][name] for name in test[1].names()])
            x_test = data_test[:-1].T.astype('float64')
            y_test = data_test[-1]
        except Exception:
            x_train, y_train = self.load_re_arff(temp_data_path + dataset_name + '_TRAIN.arff')
            x_test, y_test = self.load_re_arff(temp_data_path + dataset_name + '_TEST.arff')
        return x_test, x_train, y_test, y_train

    @staticmethod
    def read_txt(dataset_name: str, temp_data_path: str):
        """
        Reads data from ``.txt`` file.

        Args:
            dataset_name: name of dataset
            temp_data_path: path to temporary folder with downloaded data

        Returns:
            train and test data tuple
        """
        data_train = np.genfromtxt(temp_data_path + '/' + dataset_name + '_TRAIN.txt')
        data_test = np.genfromtxt(temp_data_path + '/' + dataset_name + '_TEST.txt')
        x_train, y_train = data_train[:, 1:], data_train[:, 0]
        x_test, y_test = data_test[:, 1:], data_test[:, 0]
        return x_test, x_train, y_test, y_train

    @staticmethod
    def read_tsv(file_name: str) -> tuple:
        """Read ``tsv`` file that contains data for classification experiment. Data must be placed
        in ``data`` folder with ``.tsv`` extension.

        Args:
            file_name: name of file with data

        Returns:
            tuple: (x_train, x_test) and (y_train, y_test)

        """
        try:
            df_train = pd.read_csv(os.path.join(PROJECT_PATH, 'data', file_name, f'{file_name}_TRAIN.tsv'),
                                   sep='\t',
                                   header=None)

            x_train = df_train.iloc[:, 1:]
            y_train = df_train[0].values

            df_test = pd.read_csv(os.path.join(PROJECT_PATH, 'data', file_name, f'{file_name}_TEST.tsv'),
                                  sep='\t',
                                  header=None)

            x_test = df_test.iloc[:, 1:]
            y_test = df_test[0].values
            try:
                y_train, y_test = y_train.astype(int), y_test.astype(int)
            except ValueError:
                y_train, y_test = y_train.astype(str), y_test.astype(str)

            return (x_train, y_train), (x_test, y_test)
        except FileNotFoundError:
            return None, None

    @staticmethod
    def read_tstime(file_name: str) -> tuple:
        """Read ``tsv`` file that contains data for classification experiment. Data must be placed
        in ``data`` folder with ``.tsv`` extension.

        Args:
            file_name: name of file with data

        Returns:
            tuple: (x_train, x_test) and (y_train, y_test)

        """
        try:

            x_test, y_test = load_from_tsfile_to_dataframe(os.path.join(PROJECT_PATH, 'data', file_name, f'{file_name}_TEST.ts'),
                                                           return_separate_X_and_y=True)
            x_train, y_train = load_from_tsfile_to_dataframe(os.path.join(PROJECT_PATH, 'data', file_name, f'{file_name}_TRAIN.ts'),
                                                             return_separate_X_and_y=True)
            try:
                y_train, y_test = y_train.astype(int), y_test.astype(int)
            except ValueError:
                y_train, y_test = y_train.astype(str), y_test.astype(str)

            return (x_train, y_train), (x_test, y_test)
        except FileNotFoundError:
            return None, None
    @staticmethod
    def load_re_arff(data):
        """
        Reads data from relational ``.arff`` file.

        Args:
            data: ...

        Returns:

        """
        x_data = np.asarray(data[0])
        n_samples = len(x_data)
        x_ls, y_ls = [], []

        if x_data[0][0].dtype.names is None:
            for i in range(n_samples):
                x_sample = np.asarray([x_data[i][name] for name in x_data[i].dtype.names])
                x_ls.append(x_sample.T)
                y_ls.append(x_data[i][1])
        else:
            for i in range(n_samples):
                x_sample = np.asarray([x_data[i][0][name] for name in x_data[i][0].dtype.names])
                x_ls.append(x_sample.T)
                y_ls.append(x_data[i][1])

        x_arr = np.asarray(x_ls).astype('float64')
        y_arr = np.asarray(y_ls)

        try:
            y_arr = y_arr.astype('float64').astype('int64')
        except ValueError:
            y_arr = y_arr.astype(str)

        return x_arr, y_arr



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error

from fedot_ind.core.metrics.loss.basis_loss import basis_approximation_metric
from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation


class LegenderBasis(BasisDecompositionImplementation):

    def _get_basis(self, data, n_components: int = None):
        self.basis = []
        self.coefs = []
        self.rmse = []
        domain = np.arange(self.data_range)
        for i in range(1, self.n_components + 1):
            legendre_coefs = np.polynomial.legendre.legfit(x=domain, y=data, deg=i)
            new_y = np.polynomial.legendre.legval(domain, legendre_coefs)
            self.coefs.append(legendre_coefs)
            self.basis.append(new_y)
            self.rmse.append(mean_squared_error(y_true=data,
                                                y_pred=new_y,
                                                squared=False))

    def fit(self, data):
        # Create the orthogonal polynomials
        self._get_basis(data, self.n_components)
        self.data = data

    def evaluate_derivative(self, order):
        domain = np.arange(self.data_range)
        self.derv_coef = []
        for poly_coef in self.coefs:
            derivative_of_poly = np.polynomial.legendre.legder(poly_coef, order)
            new_y = np.polynomial.legendre.legval(domain, derivative_of_poly)
            self.derv_coef.append(new_y)

    def analytical_form(self):
        polynom_number = basis_approximation_metric(metric_values=np.array(self.rmse),
                                                    derivation_coef=self.derv_coef,
                                                    regularization_coef=0.1)
        polynom_coef = self.coefs[polynom_number]
        p = np.polynomial.legendre.leg2poly(polynom_coef)
        anylytical_form = ''
        for index, coef in enumerate(p):
            degree = f' +{round(coef, 3)}*x^({len(p) - index - 1}) '
            anylytical_form = anylytical_form + degree
        anylytical_form = anylytical_form.replace('+-', '-')
        print(f'Best degree of Legendre Polynomial is - {polynom_number + 1}')
        print(anylytical_form)
        return polynom_number

    def show(self, visualisation_type: str = 'basis representation', **kwargs):
        if visualisation_type == 'basis representation':
            A = np.array(self.basis).T
        else:
            A = np.array(self.derv_coef).T
        A = pd.DataFrame(A)
        A['original'] = self.data
        A[[kwargs['basis_function'], 'original']].plot()
        plt.show()

    # def transform(self, X):
    #     # Create the orthogonal polynomials
    #
    #
    #     # Create the matrix of polynomials
    #     A = np.array(polys).T
    #
    #     # Transform the data using the coefficients
    #     return A.dot(self.coefficients)



import numpy as np
import pandas as pd

import logging
from fedot_ind.core.architecture.postprocessing.Analyzer import PerformanceAnalyzer
from fedot_ind.core.models.ecm.error_correction import Booster


class ErrorCorrectionModel:
    """Error correction model (ECM) for time series classification predictions.
    It is based on the idea of boosting: the model is trained on the residuals of the previous model predictions
    and the final prediction is the sum of the predictions of all models or another Fedot AutoML pipeline.

    Args:
        results_on_train: np.ndarray with proba_predictions on train data
        results_on_test: dictionary with predictions, proba_predictions and metrics on test data
        n_classes: number of classes
        dataset_name: name of dataset
        save_models: flag for saving ECM models
        fedot_params: dictionary with parameters for Fedot AutoML model
        train_data: train features and target
        test_data: test features and target

    """

    def __init__(self, results_on_train: np.ndarray = None,
                 results_on_test: dict = None, n_classes=None,
                 dataset_name: str = None, save_models: bool = False,
                 fedot_params: dict = None, train_data=None, test_data=None):

        self.logger = logging.getLogger(self.__class__.__name__)
        self.results_on_train = results_on_train
        self.results_on_test = results_on_test
        self.n_classes = n_classes
        self.dataset_name = dataset_name
        self.save_models = save_models
        self.fedot_params = fedot_params
        self.train_feats, self.test_features = train_data[0], test_data[0]
        self.y_train, self.y_test = train_data[1], test_data[1]
        self.analyzer = PerformanceAnalyzer()

    def run(self):
        self.logger.info('Start error correction')
        self._get_dimension_params(predictions_proba_train=self.results_on_train)

        boosting_results = self._predict_with_boosting(predictions_proba=self.results_on_test['predictions_proba'],
                                                       metrics_without_boosting=self.results_on_test['metrics'])

        return boosting_results

    def _get_dimension_params(self, predictions_proba_train: np.ndarray) -> None:
        """Creates base predict for boosting and reshape flag for predictions.

        Args:
            predictions_proba_train: array with proba_predictions on train data

        """
        if self.n_classes > 2:
            self.base_predict = predictions_proba_train
            y_train_multi = pd.get_dummies(self.y_train, sparse=True)
            self.y_train = y_train_multi.values
            self.reshape_flag = False
        else:
            self.base_predict = predictions_proba_train.reshape(-1)
            self.base_predict = round(np.max(self.y_test) - np.max(predictions_proba_train)) + self.base_predict
            self.reshape_flag = True

    def _predict_with_boosting(self, predictions_proba: np.ndarray,
                               metrics_without_boosting: dict) -> dict:
        """Instantiates the Booster class and predicts on test features using boosting pipeline consists of
        several models and ensemble method.

        Args:
            predictions_proba: np.ndarray with proba_predictions on test data
            metrics_without_boosting: dictionary with metrics on test data without boosting

        Returns:
            boosting_results (dict): dictionary with predictions, proba_predictions
            and metrics on test data with boosting

        """
        booster = Booster(features_train=self.train_feats,
                          target_train=self.y_train,
                          base_predict=self.base_predict,
                          timeout=1,
                          reshape_flag=self.reshape_flag)

        boosting_pipeline = self.genetic_boosting_pipeline

        _, model_list, ensemble_model = booster.fit()
        results_on_test = boosting_pipeline(self.results_on_test['prediction'],
                                            model_list,
                                            ensemble_model)

        results_on_test['base_probs_on_test'] = predictions_proba

        proba_to_vector = booster.proba_to_vector

        solution_table = pd.DataFrame({'target': self.y_test,
                                       'base_probs_on_test': proba_to_vector(results_on_test['base_probs_on_test']),
                                       'corrected_probs': proba_to_vector(results_on_test['corrected_probs']),
                                       'corrected_labels': proba_to_vector(results_on_test['corrected_labels'])})

        self.metrics_name = ['f1', 'roc_auc', 'accuracy', 'logloss', 'precision']
        metrics_boosting = self.analyzer.calculate_metrics(metric_list=self.metrics_name,
                                                           target=self.y_test,
                                                           predicted_labels=results_on_test['corrected_labels'],
                                                           predicted_probs=results_on_test['corrected_probs'])

        no_boost = pd.Series(metrics_without_boosting)
        boost = pd.Series(metrics_boosting)
        metrics_table = pd.concat([no_boost, boost], axis=1).reset_index()
        metrics_table.columns = ['metric', 'before_boosting', 'after_boosting']

        return dict(solution_table=solution_table,
                    metrics_table=metrics_table,
                    model_list=model_list,
                    ensemble_model=ensemble_model)

    def genetic_boosting_pipeline(self, predictions_proba, model_list, ensemble_model) -> dict:
        self.logger.info('Predict on booster models')
        boosting_stages_predict = []
        input_data_test = self.test_features

        n = 1
        for model in model_list:
            self.logger.info(f'Cycle {n} of boosting has started')
            boost_predict = model.predict(input_data_test)
            boosting_stages_predict.append(boost_predict)
            n += 1

        self.logger.info('Ensemble booster predictions')
        if ensemble_model:
            self.logger.info('Ensemble using FEDOT has been chosen')
            boosting_stages_predict = pd.DataFrame(i.reshape(-1) for i in boosting_stages_predict).T
        else:
            boosting_stages_predict = [np.array(_) for _ in boosting_stages_predict]
            self.logger.info('Ensemble using SUM method has been chosen')

        boosting_result = self._convert_boosting_prediction(boosting_stages_predict=boosting_stages_predict,
                                                            ensemble_model=ensemble_model,
                                                            predictions_proba=predictions_proba)
        return boosting_result

    def _convert_boosting_prediction(self, boosting_stages_predict, ensemble_model, predictions_proba):
        self.logger.info('Calculation of error correction by boosting')
        # if self.static_booster:
        #     task = Task(TaskTypesEnum.regression)
        #     input_data = InputData(idx=np.arange(0, len(boosting_stages_predict)),
        #                            features=boosting_stages_predict,
        #                            target=self.y_test,
        #                            task=task,
        #                            data_type=DataTypesEnum.table)
        # else:
        input_data = boosting_stages_predict

        if self.reshape_flag:
            error_correction = ensemble_model.predict(input_data).reshape(-1)
            corrected_probs = error_correction.reshape(-1) + predictions_proba.reshape(-1)

            if self.reshape_flag and np.min(self.y_test) == 1:
                corrected_probs = (error_correction.reshape(-1) + predictions_proba.reshape(-1)) / 2
            corrected_probs = corrected_probs.reshape(-1)
            corrected_labels = abs(np.round(corrected_probs))

        else:
            dictlist = []
            for value in boosting_stages_predict:
                dictlist.append(value)
            error_correction = sum(dictlist)
            corrected_probs = error_correction + predictions_proba
            corrected_labels = np.array([x.argmax() + min(self.y_test) for x in corrected_probs])
            corrected_labels = corrected_labels.reshape(len(corrected_labels), 1)

        return dict(corrected_labels=corrected_labels,
                    corrected_probs=corrected_probs,
                    error_correction=error_correction)



import os
from typing import Optional

import numpy as np
from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import \
    DataOperationImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum

from fedot_ind.core.architecture.utils.utils import PROJECT_PATH
from fedot_ind.core.operation.utils.cache import DataCacher


class IndustrialCachableOperationImplementation(DataOperationImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)

        cache_folder = os.path.join(PROJECT_PATH, 'cache')
        os.makedirs(cache_folder, exist_ok=True)
        self.cacher = DataCacher(data_type_prefix=f'Features of basis',
                                 cache_folder=cache_folder)

        self.data_type = DataTypesEnum.image

    def fit(self, data):
        """Decomposes the given data on the chosen basis.

        Returns:
            np.array: The decomposition of the given data.
        """
        pass

    def try_load_from_cache(self, hashed_info: str) -> np.array:
        predict = self.cacher.load_data_from_cache(hashed_info=hashed_info)
        return predict

    def transform(self, input_data: InputData) -> OutputData:
        """
            Method firstly tries to load result from cache. If unsuccessful, it starts to generate features
        """

        hashed_info = self.cacher.hash_info(data=input_data.features.tobytes(),
                                            generator_info=self.__dir__())
        try:
            predict = self.try_load_from_cache(hashed_info)
        except FileNotFoundError:
            predict = self._transform(input_data)
            self.cacher.cache_data(hashed_info, predict)

        predict = self._convert_to_output(input_data, predict, data_type=self.data_type)
        return predict

    def _transform(self, input_data) -> np.array:
        """
            Method for feature generation for all series
        """
        pass



from multiprocessing import cpu_count
from typing import Optional
import logging

from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from tqdm import tqdm

from fedot_ind.core.metrics.metrics_implementation import *
from fedot_ind.core.operation.IndustrialCachableOperation import IndustrialCachableOperationImplementation
from fedot_ind.core.operation.utils.cache import DataCacher


class BaseExtractor(IndustrialCachableOperationImplementation):
    """
    Abstract class responsible for feature generators.
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.current_window = None
        self.n_processes = cpu_count() // 2
        self.data_type = DataTypesEnum.table
        self.use_cache = params.get('use_cache', False)

        self.logger = logging.getLogger(self.__class__.__name__)

    def fit(self, input_data: InputData):
        pass

    def _transform(self, input_data: InputData) -> np.array:
        """
            Method for feature generation for all series
        """
        v = []

        for series in tqdm(np.squeeze(input_data.features, 3)):
            v.append(self.generate_features_from_ts(series))
        predict = self._clean_predict(np.array(v))

        return predict

    @staticmethod
    def _clean_predict(predict: np.array):
        """
            Clean predict from nan, inf and reshape data for Fedot appropriate form
        """
        predict = np.where(np.isnan(predict), 0, predict)
        predict = np.where(np.isinf(predict), 0, predict)
        predict = predict.reshape(predict.shape[0], -1)
        return predict

    def generate_features_from_ts(self, ts_frame: pd.DataFrame,
                                  window_length: int = None) -> pd.DataFrame:
        """Method responsible for generation of features from time series.
        """
        pass

    def extract_features(self, train_features: pd.DataFrame,
                         dataset_name: str = None) -> pd.DataFrame:
        """Wrapper method for feature extraction method get_features() with caching results into pickle file. The idea
        is to create a unique pointer from dataset name, subsample (test or train) and feature generator object. We
        can uniquely identify the generator in our case only using a set of parameters in the form of obj.__dict__,
        while excluding some dynamic attributes. In this way we can create a hash of incoming data unique for each
        case, and then associate it with the output data - the feature set.

        Args:
            train_features: dataframe with time series.
            dataset_name: name of dataset.

        Returns:
            Dataframe with extracted features.

        """
        generator_name = self.__class__.__name__

        if self.use_cache:
            generator_info = self.__dir__()
            data_cacher = DataCacher()
            hashed_info = data_cacher.hash_info(data=train_features,
                                                name=dataset_name,
                                                obj_info_dict=generator_info)
            try:
                return data_cacher.load_data_from_cache(hashed_info)
            except FileNotFoundError:
                self.logger.info('Cache not found. Generating features')

                features = self.generate_features_from_ts(train_features, dataset_name)
                data_cacher.cache_data(hashed_info, features)
                return features
        else:
            return self.generate_features_from_ts(train_features, dataset_name)



import numpy as np
from sklearn.preprocessing import MinMaxScaler


def sv_to_explained_variance_ratio(singular_values, rank):
    """
    Calculate the explained variance ratio of the singular values.

    Parameters
    ----------
    singular_values : array-like, shape (n_components,)
        Singular values.
    rank : int
        Number of singular values to use.

    Returns
    -------
    explained_variance : float
        Explained variance ratio.
    n_components : int
        Number of singular values to use.
    """
    n_components = [x / sum(singular_values) * 100 for x in singular_values][:rank]
    explained_variance = sum(n_components)
    n_components = rank
    return explained_variance, n_components


def singular_value_hard_threshold(singular_values, rank=None, beta=None, threshold=2.858) -> list:
    """
    Calculate the hard threshold for the singular values.

    Parameters
    ----------
    singular_values : array-like, shape (n_components,)
        Singular values.
    rank : int
        Number of singular values to use.
    threshold : float
        Threshold value.

    Returns
    -------
    adjusted_rank : int
        Adjusted rank.
    """
    if rank is not None:
        return singular_values[:rank]
    else:
        # Scale the singular values between 0 and 1.
        singular_values_scaled = abs(singular_values)
        singular_values_scaled = MinMaxScaler(feature_range=(0, 1)).fit_transform(
            singular_values_scaled.reshape(-1, 1))[:, 0]
        # Find the median of the singular values.
        median_sv = np.median(singular_values_scaled[:rank])
        # Find the adjusted rank.
        if threshold is None:
            threshold = 0.56 * np.power(beta, 3) - 0.95 * np.power(beta, 2) + 1.82 * beta + 1.43
        sv_threshold = threshold * median_sv
        # Find the threshold value.
        adjusted_rank = np.sum(singular_values_scaled >= sv_threshold)
        # If the adjusted rank is 0 or 1, set it to 2.
        if adjusted_rank < 2:
            adjusted_rank = 2
        return singular_values[:adjusted_rank]


def reconstruct_basis(U, Sigma, VT, ts_length):
    if len(Sigma.shape) > 1:
        multi_reconstruction = lambda x: reconstruct_basis(U=U, Sigma=x, VT=VT, ts_length=ts_length)
        TS_comps = list(map(multi_reconstruction, Sigma))
    else:
        rank = Sigma.shape[0]
        TS_comps = np.zeros((ts_length, rank))
        for i in range(rank):
            X_elem = Sigma[i] * np.outer(U[:, i], VT[i, :])
            X_rev = X_elem[::-1]
            eigenvector = [X_rev.diagonal(j).mean() for j in range(-X_rev.shape[0] + 1, X_rev.shape[1])]
            TS_comps[:, i] = eigenvector
    return TS_comps



from enum import Enum

from fedot_ind.core.models.detection.probalistic.kalman import UnscentedKalmanFilter
from fedot_ind.core.models.detection.subspaces.func_pca import FunctionalPCA
from fedot_ind.core.models.detection.subspaces.sst import SingularSpectrumTransformation
from fedot_ind.core.models.signal.RecurrenceExtractor import RecurrenceExtractor
from fedot_ind.core.models.signal.SignalExtractor import SignalExtractor
from fedot_ind.core.models.statistical.StatsExtractor import StatsExtractor
from fedot_ind.core.models.topological.TopologicalExtractor import TopologicalExtractor
from fedot_ind.core.operation.transformation.basis.chebyshev import ChebyshevBasis
from fedot_ind.core.operation.transformation.basis.data_driven import DataDrivenBasisImplementation
from fedot_ind.core.operation.transformation.basis.fourier import FourierBasisImplementation
from fedot_ind.core.operation.transformation.basis.wavelet import WaveletBasisImplementation


class BasisTransformations(Enum):
    chebyshev = ChebyshevBasis
    datadriven = DataDrivenBasisImplementation
    wavelet = WaveletBasisImplementation
    Fourier = FourierBasisImplementation


class FeatureGenerator(Enum):
    statistical = StatsExtractor
    wavelet = SignalExtractor
    topological = TopologicalExtractor
    recurrence = RecurrenceExtractor


class MlModel(Enum):
    functional_pca = FunctionalPCA
    kalman_filter = UnscentedKalmanFilter
    sst = SingularSpectrumTransformation


class KernelFeatureGenerator(Enum):
    statistical = [{'feature_generator_type': 'statistical',
                    'feature_hyperparams': {
                        'window_mode': True,
                        'window_size': 5
                    }
                    },
                   {'feature_generator_type': 'statistical',
                    'feature_hyperparams': {
                        'window_mode': True,
                        'window_size': 10
                    }
                    },
                   {'feature_generator_type': 'statistical',
                    'feature_hyperparams': {
                        'window_mode': True,
                        'window_size': 20
                    }
                    },
                   {'feature_generator_type': 'statistical',
                    'feature_hyperparams': {
                        'window_mode': True,
                        'window_size': 30
                    }
                    },
                   {'feature_generator_type': 'statistical',
                    'feature_hyperparams': {
                        'window_mode': True,
                        'window_size': 40
                    }
                    }
                   ]
    wavelet = [
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "mexh",
             'n_components': 2
         }},
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "haar",
             'n_components': 2
         }
         },
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "dmey",
             'n_components': 2
         }
         },
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "gaus3",
             'n_components': 2
         }
         },
        {'feature_generator_type': 'wavelet',
         'feature_hyperparams': {
             'wavelet': "morl",
             'n_components': 2
         }
         }
    ]
    recurrence = []
    topological = []



from functools import partial
import pandas as pd
from pymonad.list import ListMonad
from pymonad.either import Right
from fedot_ind.core.architecture.pipelines.abstract_pipeline import AbstractPipelines


class AnomalyDetectionPipelines(AbstractPipelines):

    def __call__(self, pipeline_type: str = 'SpecifiedFeatureGeneratorTSC'):
        pipeline_dict = {'SST': self.__singular_transformation_pipeline,
                         'FunctionalPCA': self.__functional_pca_pipeline,
                         'Kalman': self.__kalman_filter_pipeline,
                         'VectorAngle': self.__vector_based_pipeline,
                         }
        return pipeline_dict[pipeline_type]

    def get_feature_generator(self, **kwargs):
        lambda_func_dict = kwargs['steps']
        lambda_func_dict['concatenate_features'] = lambda x: list_of_features.append(pd.DataFrame(x))
        input_data = kwargs['input_data']
        list_of_features = []
        if kwargs['pipeline_type'] == 'SST':
            pipeline = Right(input_data).then(lambda_func_dict['fit_model'])
        elif kwargs['pipeline_type'] == 'FunctionalPCA':
            if type(input_data) != pd.DataFrame:
                input_data = pd.DataFrame(input_data)
            pipeline = Right(input_data).then(lambda_func_dict['create_list_of_ts']).map(
                lambda_func_dict['transform_to_basis']).then(lambda_func_dict['reduce_basis']).then(
                lambda_func_dict['concatenate_features']).insert(
                self._get_feature_matrix(list_of_features=list_of_features, mode='1D'))
        elif kwargs['pipeline_type'] == 'Kalman':
            pipeline = Right(input_data).then(lambda_func_dict['fit_model']).insert(self._get_feature_matrix(
                list_of_features=self.test_features, mode='list_of_ts', window_length=kwargs['window_length']))
        elif kwargs['pipeline_type'] == 'VectorAngle':
            pipeline = Right(input_data)
        return pipeline

    def __singular_transformation_pipeline(self, mode: str = 'multits', **kwargs):
        feature_extractor, detector, lambda_func_dict = self._init_pipeline_nodes(model_type='sst',
                                                                                  **kwargs)

        if mode == 'multits':
            self.test_features = [pd.DataFrame(time_series) for time_series in self.test_features]
            self.test_features = [time_series.values.tolist() for time_series in self.test_features]

        lambda_func_dict['fit_model'] = lambda list_of_ts: [detector.fit(train_features=slice_ts)
                                                            for slice_ts in list_of_ts]
        lambda_func_dict['predict'] = lambda list_of_ts: [detector.predict(test_features=slice_ts[0],target=slice_ts[1])
                                                          for slice_ts in list_of_ts]

        train_feature_generator_module = self.get_feature_generator(input_data=self.train_features,
                                                                    steps=lambda_func_dict,
                                                                    pipeline_type='SST'
                                                                    )
        prediction = train_feature_generator_module.map(lambda_func_dict['predict'])

        detector.singular_transformation = partial(self.get_feature_generator,
                                                   steps=lambda_func_dict,
                                                   pipeline_type='SST')
        return detector, prediction.value

    def __multi_functional_pca(self):
        pass

    def __functional_pca_pipeline(self, decomposition: str = 'svd', **kwargs):
        feature_extractor, detector, lambda_func_dict = self._init_pipeline_nodes(model_type='functional_pca',
                                                                                  **kwargs)
        data_basis = DataDrivenBasis()
        if kwargs['component'] is None:
            kwargs['component'] = [0 for i in range(self.train_features.shape[1])]
        if decomposition == 'tensor':
            lambda_func_dict['create_list_of_ts'] = lambda time_series: ListMonad(time_series.values.tolist())
        else:
            lambda_func_dict['create_list_of_ts'] = lambda time_series: ListMonad(*time_series.values.tolist())

        lambda_func_dict['transform_to_basis'] = lambda time_series: ListMonad(self.basis if self.basis \
                                                                                             is not None else data_basis.fit(
            time_series, window_length=None))
        lambda_func_dict['reduce_basis'] = lambda list_of_components: [component[:, rank] for component, rank in
                                                                       zip(list_of_components, kwargs['component'])]
        lambda_func_dict['predict'] = lambda list_of_ts: [detector.predict(test_features=slice_ts)
                                                          for slice_ts in list_of_ts]

        train_feature_generator_module = self.get_feature_generator(input_data=self.train_features,
                                                                    steps=lambda_func_dict,
                                                                    pipeline_type='FunctionalPCA')
        detector.fit(train_feature_generator_module.value[0])

        prediction = train_feature_generator_module.insert(self.test_features).map(lambda_func_dict['predict'])
        detector.feature_generator = partial(self.get_feature_generator, steps=lambda_func_dict,
                                             pipeline_type='FunctionalPCA')

        return detector, prediction.value[0]

    def __kalman_filter_pipeline(self, **kwargs):

        feature_extractor, detector, lambda_func_dict = self._init_pipeline_nodes(model_type='kalman_filter',
                                                                                  **kwargs)
        lambda_func_dict['fit_model'] = lambda train_features: detector.fit(train_features=train_features)
        lambda_func_dict['predict'] = lambda list_of_ts: [detector.predict(test_features=slice_ts)
                                                          for slice_ts in list_of_ts]

        train_feature_generator_module = self.get_feature_generator(input_data=self.train_features,
                                                                    steps=lambda_func_dict,
                                                                    pipeline_type='Kalman',
                                                                    window_length=kwargs['window_length']
                                                                    )
        prediction = train_feature_generator_module.map(lambda_func_dict['predict'])

        detector.state_transition_matrix = partial(self.get_feature_generator,
                                                   steps=lambda_func_dict,
                                                   pipeline_type='Kalman',
                                                   window_length=kwargs['window_length'])
        return detector, prediction.value

    def __vector_based_pipeline(self):
        pass



# -*- coding: utf-8 -*-
from typing import Union

import numpy as np
import pandas as pd

from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix
from fedot_ind.core.operation.transformation.regularization.spectrum import singular_value_hard_threshold

supported_types = (pd.Series, np.ndarray, list)


class SpectrumDecomposer:
    """Decomposes the given time series with a singular-spectrum analysis. Assumes the values of the time series are
    recorded at equal intervals.

    Args:
        time_series: The time series to decompose.
        window_length: The length of the window to use. Defaults to None.
        save_memory: Whether to save memory by not storing the elementary matrices. Defaults to True.

    """

    def __init__(self,
                 time_series: Union[pd.DataFrame, pd.Series, np.ndarray, list],
                 window_length: int = None,
                 save_memory: bool = True):
        self.__time_series = time_series
        self.__convert_ts_to_array()
        self.__window_length = window_length
        if self.__window_length is None:
            self.__window_length = round(self.__time_series.size * 0.35)
        self.__save_memory = save_memory
        self.__set_dimensions()
        self.__check_windows_length()
        self.__trajectory_matrix = self.__get_trajectory_matrix()

    def __check_windows_length(self):
        if not 2 <= self.__window_length <= self.__ts_length / 2:
            raise ValueError("The window length must be in the interval [2, N/2].")

    def __convert_ts_to_array(self):
        if type(self.__time_series) == pd.DataFrame:
            self.__time_series = self.__time_series.values
        elif type(self.__time_series) == list:
            self.__time_series = np.array(self.__time_series)
        else:
            self.__time_series = self.__time_series

    def __set_dimensions(self):
        self.__ts_length = len(self.__time_series)
        self.__subseq_length = self.__ts_length - self.__window_length + 1

    def __get_trajectory_matrix(self):
        # return np.array(
        #     [self.__time_series[i:self.__window_length + i] for i in range(0, self.__subseq_length)]).T
        hankel_transform = HankelMatrix(time_series=self.__time_series, window_size=self.window_length)
        if len(self.__time_series.shape) > 2:
            return hankel_transform.trajectory_matrix
        else:
            return hankel_transform.trajectory_matrix

    @property
    def window_length(self):
        return self.__window_length

    @property
    def sub_seq_length(self):
        return self.__subseq_length

    @window_length.setter
    def window_length(self, window_length):
        self.__window_length = window_length

    @property
    def trajectory_matrix(self):
        return self.__trajectory_matrix

    @property
    def ts_length(self):
        return self.__ts_length

    @trajectory_matrix.setter
    def trajectory_matrix(self, trajectory_matrix: np.ndarray):
        self.__trajectory_matrix = trajectory_matrix

    def decompose(self, rank_hyper: int = None):
        # Embed the time series in a trajectory matrix
        U, Sigma, VT = np.linalg.svd(self.__trajectory_matrix)

        if rank_hyper is None:
            sing_values = singular_value_hard_threshold(singular_values=Sigma)
            rank = len(sing_values)

        else:
            rank = rank_hyper

        # Decompose the trajectory matrix
        TS_comps = np.zeros((self.__ts_length, rank))

        if not self.__save_memory:
            # Construct and save all the elementary matrices
            X_elem = np.array([Sigma[i] * np.outer(U[:, i], VT[i, :]) for i in range(rank)])

            # Diagonally average the elementary matrices, store them as columns in array.
            for i in range(rank):
                X_rev = X_elem[i, ::-1]
                TS_comps[:, i] = [X_rev.diagonal(j).mean() for j in range(-X_rev.shape[0] + 1, X_rev.shape[1])]

            V = VT.T
        else:
            # Reconstruct the elementary matrices without storing them
            for i in range(rank):
                X_elem = Sigma[i] * np.outer(U[:, i], VT[i, :])
                X_rev = X_elem[::-1]
                TS_comps[:, i] = [X_rev.diagonal(j).mean() for j in range(-X_rev.shape[0] + 1, X_rev.shape[1])]

            X_elem = "Re-run with save_mem=False to retain the elementary matrices."

            # The V array may also be very large under these circumstances, so we won't keep it.
            V = "Re-run with save_mem=False to retain the V matrix."

        return TS_comps, Sigma, rank, X_elem, V

    @staticmethod
    def components_to_df(TS_comps: np.ndarray, rank: int, n: int = 0) -> pd.DataFrame:
        """Converts all the time series components in a single Pandas DataFrame object.

        Args:
            TS_comps: The time series components.
            rank: The rank of the time series.
            n: ...

        Returns:
            df: dataframe with all the time series components.
        """
        if n > 0:
            n = min(n, rank)
        else:
            n = rank

        # Create list of columns - call them F0, F1, F2, ...
        cols = ["F{}".format(i) for i in range(n)]
        df = pd.DataFrame(TS_comps.T).T.iloc[:, :rank]
        df.columns = cols
        return df

    def reconstruct(self, indices: Union[int, list, tuple, np.ndarray]) -> pd.Series:
        """Reconstructs the time series from its elementary components, using the given indices. Returns a Pandas Series
        object with the reconstructed time series.

        Args:
            indices: An integer, list of integers or slice(n,m) object, representing the elementary
            components to sum.

        Returns:
            The reconstructed time series.
        """
        if isinstance(indices, int): indices = [indices]

        ts_vals = self.TS_comps[:, indices].sum(axis=1)
        return pd.Series(ts_vals)



import logging
import os
from typing import List, Union
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import torch
from fedot.api.main import Fedot
from fedot.core.data.data import array_to_input_data
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline

from fedot_ind.api.utils.checkers_collections import DataCheck
from fedot_ind.api.utils.saver_collections import ResultSaver
from fedot_ind.core.architecture.datasets.classification_datasets import CustomClassificationDataset
from fedot_ind.core.architecture.experiment.CVModule import ClassificationExperimenter
from fedot_ind.core.architecture.postprocessing.Analyzer import PerformanceAnalyzer
from fedot_ind.core.architecture.utils.utils import default_path_to_save_results
from fedot_ind.core.models.nn.inception import InceptionTimeNetwork

TSCCLF_MODEL = {
    'inception_time': InceptionTimeNetwork
}


class TimeSeriesClassifier:
    """Class responsible for interaction with Fedot classifier.

    Args:
        params: which are ``generator_name``, ``generator_runner``, ``model_hyperparams``,
                ``ecm_model_flag``, ``dataset_name``, ``output_dir``.


    Attributes:
        logger (Logger): logger instance
        predictor (Fedot): Fedot model instance
        y_train (np.ndarray): target for training
        train_features (pd.DataFrame): features for training

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        self.generator_name = params.get('feature_generator', 'statistical')
        self.model_hyperparams = params.get('model_params')
        self.generator_runner = params.get('generator_class')
        self.dataset_name = params.get('dataset')
        self.output_folder = params.get('output_folder', None)

        self.saver = ResultSaver(dataset_name=self.dataset_name,
                                 generator_name=self.generator_name,
                                 output_dir=self.output_folder)
        self.logger = logging.getLogger('TimeSeriesClassifier')
        self.datacheck = DataCheck()

        self.prediction_label = None
        self.predictor = None
        self.y_train = None
        self.train_features = None
        self.test_features = None
        self.input_test_data = None

        self.logger.info('TimeSeriesClassifier initialised')

    def _fit_model(self, features: pd.DataFrame, target: np.ndarray) -> Fedot:
        """Fit Fedot model with feature and target.

        Args:
            features: features for training
            target: target for training

        Returns:
            Fitted Fedot model

        """
        self.predictor = Fedot(**self.model_hyperparams)
        self.predictor.fit(features, target)
        return self.predictor

    def _fit_baseline_model(self, features: pd.DataFrame, target: np.ndarray, baseline_type: str = 'rf') -> Pipeline:
        """Returns pipeline with the following structure:
            ``[initial data] -> [scaling] -> [baseline type]``

        By default, baseline type is random forest, but it can be changed to any other model from Fedot library:
        logit, knn, svc, qda, xgboost.

        Args:
            features: features for training
            target: target for training
            baseline_type: type of baseline model

        Returns:
            Fitted Fedot pipeline with baseline model

        """
        self.logger.info(f'Baseline model pipeline: scaling -> {baseline_type}. Fitting...')
        node_scaling = PrimaryNode('scaling')
        node_final = SecondaryNode(baseline_type,
                                   nodes_from=[node_scaling])
        baseline_pipeline = Pipeline(node_final)
        input_data = array_to_input_data(features_array=features,
                                         target_array=target)
        baseline_pipeline.fit(input_data)
        self.logger.info(f'Baseline model has been fitted')
        return baseline_pipeline

    def __predict_abstraction(self,
                              test_features: Union[np.ndarray, pd.DataFrame],
                              mode: str = 'labels'):
        self.logger.info(f'Predicting with {self.generator_name} generator')

        if self.test_features is None:
            self.test_features = self.generator_runner.extract_features(train_features=test_features,
                                                                        dataset_name=self.dataset_name)
            self.test_features = self.datacheck.check_data(input_data=self.test_features, return_df=True)

        if isinstance(self.predictor, Pipeline):
            self.input_test_data = array_to_input_data(features_array=self.test_features, target_array=None)
            prediction_label = self.predictor.predict(self.input_test_data, output_mode=mode).predict
            return prediction_label
        else:
            if mode == 'labels':
                prediction_label = self.predictor.predict(self.test_features)
            else:
                prediction_label = self.predictor.predict_proba(self.test_features)
            return prediction_label

    def fit(self, train_ts_frame: Union[np.ndarray, pd.DataFrame],
            train_target: np.ndarray,
            **kwargs) -> object:

        baseline_type = kwargs.get('baseline_type', None)
        self.logger.info(f'Fitting model')
        self.y_train = train_target
        if self.generator_runner is None:
            raise AttributeError('Feature generator is not defined')

        train_features = self.generator_runner.extract_features(train_features=train_ts_frame,
                                                                dataset_name=self.dataset_name)

        self.train_features = self.datacheck.check_data(input_data=train_features,
                                                        return_df=True)

        if baseline_type is not None:
            self.predictor = self._fit_baseline_model(self.train_features, train_target, baseline_type)
        else:
            self.predictor = self._fit_model(self.train_features, train_target)

        self.logger.info(
            f'Solver fitted: {self.generator_name}_extractor -> fedot_pipeline ({self.predictor.current_pipeline})')
        return self.predictor

    def predict(self, test_features: np.ndarray, **kwargs) -> dict:
        self.prediction_label = self.__predict_abstraction(test_features=test_features,
                                                           mode='labels')
        return self.prediction_label

    def predict_proba(self, test_features: np.ndarray, **kwargs) -> dict:
        self.prediction_proba = self.__predict_abstraction(test_features=test_features,
                                                           mode='probs', )
        return self.prediction_proba

    def get_metrics(self, target: Union[np.ndarray, pd.Series], metric_names: Union[str, List[str]]):
        analyzer = PerformanceAnalyzer()
        return analyzer.calculate_metrics(target=target,
                                          predicted_labels=self.prediction_label,
                                          predicted_probs=self.prediction_proba,
                                          target_metrics=metric_names)

    def save_prediction(self, predicted_data: np.ndarray, kind: str):
        self.saver.save(predicted_data, kind)

    def save_metrics(self, metrics: dict):
        self.saver.save(metrics, 'metrics')


class TimeSeriesImageClassifier(TimeSeriesClassifier):

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)

    def _init_model_param(self, target: np.ndarray) -> Tuple[int, np.ndarray]:

        num_epochs = self.model_hyperparams['epoch']
        del self.model_hyperparams['epoch']

        if 'optimization_method' in self.model_hyperparams.keys():
            modes = {'none': {},
                     'SVD': self.model_hyperparams['optimization_method']['svd_parameters'],
                     'SFP': self.model_hyperparams['optimization_method']['sfp_parameters']}
            self.model_hyperparams['structure_optimization'] = self.model_hyperparams['optimization_method']['mode']
            self.model_hyperparams['structure_optimization_params'] = modes[
                self.model_hyperparams['optimization_method']['mode']]
            del self.model_hyperparams['optimization_method']

        self.model_hyperparams['models_saving_path'] = os.path.join(default_path_to_save_results(), 'TSCImage',
                                                                    self.generator_name,
                                                                    '../../models')
        self.model_hyperparams['summary_path'] = os.path.join(default_path_to_save_results(), 'TSCImage',
                                                              self.generator_name,
                                                              'runs')
        self.model_hyperparams['num_classes'] = np.unique(target).shape[0]

        if target.min() != 0:
            target = target - 1

        return num_epochs, target

    def fit(self, train_ts_frame: Union[np.ndarray, pd.DataFrame],
            train_target: np.ndarray,
            **kwargs) -> object:
        if type(train_ts_frame) is pd.DataFrame:
            train_ts_frame =train_ts_frame.values
        return self._fit_model(features=train_ts_frame, target=train_target)

    def _fit_model(self, features: np.ndarray, target: np.ndarray) -> ClassificationExperimenter:
        """Fit Fedot model with feature and target.

        Args:
            features: features for training
            target: target for training

        Returns:
            Fitted Fedot model

        """
        num_epochs, target = self._init_model_param(target)

        train_dataset = CustomClassificationDataset(images=features, targets=target)
        NN_model = ClassificationExperimenter(train_dataset=train_dataset,
                                              val_dataset=train_dataset,
                                              **self.model_hyperparams)
        NN_model.fit(num_epochs=num_epochs)
        return NN_model

    def predict(self, test_ts_frame: np.ndarray, dataset_name: str = None) -> dict:
        prediction_label = self.__predict_abstraction(test_ts_frame, dataset_name)
        prediction_label = list(prediction_label.values())
        return dict(label=prediction_label, test_features=self.test_features)

    def predict_proba(self, test_ts_frame: np.ndarray, dataset_name: str = None) -> dict:
        prediction_proba = self.__predict_abstraction(test_features=test_ts_frame,
                                                      mode='probs')
        prediction_proba = np.concatenate(list(prediction_proba.values()), axis=0)
        return dict(class_probability=prediction_proba, test_features=self.test_features)


class TimeSeriesClassifierNN(TimeSeriesImageClassifier):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.device = torch.device('cuda' if params.get('gpu', False) else 'cpu')
        self.model = TSCCLF_MODEL[params.get('model', 'inception_time')].network_architecture
        self.num_epochs = params.get('num_epochs', 10)

    def _init_model_param(self, target: np.ndarray) -> Tuple[int, np.ndarray]:
        self.model_hyperparams['models_saving_path'] = os.path.join(default_path_to_save_results(), 'TSCNN',
                                                                    '../../models')
        self.model_hyperparams['summary_path'] = os.path.join(default_path_to_save_results(), 'TSCNN',
                                                              'runs')
        self.model_hyperparams['num_classes'] = np.unique(target).shape[0]

        if target.min() != 0:
            target = target - 1

        return self.num_epochs, target



from typing import Sequence

from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.mutation import MutationTypesEnum
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.core.optimisers.populational_optimizer import _try_unfit_graph

from fedot_ind.core.repository.IndustrialDispatcher import IndustrialDispatcher
from fedot_ind.core.repository.initializer_industrial_models import \
    has_no_data_flow_conflicts_in_industrial_pipeline


class IndustrialEvoOptimizer(EvoGraphOptimizer):
    def __init__(self,
                 objective: Objective,
                 initial_graphs: Sequence[OptGraph],
                 requirements: GraphRequirements,
                 graph_generation_params: GraphGenerationParams,
                 graph_optimizer_params: GPAlgorithmParameters):

        graph_optimizer_params.mutation_types.remove(MutationTypesEnum.single_drop)
        graph_generation_params.verifier._rules.append(has_no_data_flow_conflicts_in_industrial_pipeline)
        #graph_generation_params.verifier._rules.remove(has_no_conflicts_with_data_flow)

        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.eval_dispatcher = IndustrialDispatcher(adapter=graph_generation_params.adapter,
                                                    n_jobs=requirements.n_jobs,
                                                    graph_cleanup_fn=_try_unfit_graph,
                                                    delegate_evaluator=graph_generation_params.remote_evaluator)


import torch
from torch import Tensor
from torch.linalg import vector_norm, matrix_norm
from torch.nn.modules.module import Module


class SVDLoss(Module):
    """Base class for singular value decomposition losses.

    Args:
        factor: The hyperparameter by which the calculated loss function is multiplied
            (default: ``1``).
    """

    def __init__(self, factor: float = 1.) -> None:
        super().__init__()
        self.factor = factor


class OrthogonalLoss(SVDLoss):
    """Orthogonality regularizer for unitary matrices obtained by SVD decomposition.

    Args:
        factor: The hyperparameter by which the calculated loss function is multiplied
            (default: ``1``).
        device: The torch.device object on which the model is training.
    """

    def __init__(self, device: torch.device, factor: float = 1.) -> None:
        super().__init__(factor=factor)
        self.device = device

    def forward(self, model: Module) -> Tensor:
        """Calculates orthogonality loss.

        Args:
            model: Optimizable module containing SVD decomposed layers.
        """
        loss = 0
        n = 0
        for name, parameter in model.named_parameters():
            if name.split('.')[-1] == 'U':
                n += 1
                U = parameter
                r = U.size()[1]
                E = torch.eye(r, device=self.device)
                loss += matrix_norm(U.transpose(0, 1) @ U - E) ** 2 / r

            elif name.split('.')[-1] == 'Vh':
                Vh = parameter
                r = Vh.size()[0]
                E = torch.eye(r, device=self.device)
                loss += matrix_norm(Vh @ Vh.transpose(0, 1) - E) ** 2 / r
        return self.factor * loss / n


class HoyerLoss(SVDLoss):
    """Hoyer regularizer for matrix with singular values obtained by SVD decomposition.

        Args:
        factor: The hyperparameter by which the calculated loss function is multiplied
            (default: ``1``).
    """

    def __init__(self, factor: float = 1.) -> None:
        super().__init__(factor=factor)

    def forward(self, model: Module) -> Tensor:
        """Calculates Hoyer loss.

        Args:
            model: Optimizable module containing SVD decomposed layers.
        """
        loss = 0
        n = 0
        for name, parameter in model.named_parameters():
            if name.split('.')[-1] == 'S':
                n += 1
                S = parameter
                loss += vector_norm(S, ord=1) / vector_norm(S, ord=2)
        return self.factor * loss / n



from enum import Enum

from fedot_ind.core.architecture.experiment.TimeSeriesClassifier import TimeSeriesClassifier, TimeSeriesClassifierNN, \
    TimeSeriesImageClassifier
from fedot_ind.core.architecture.experiment.TImeSeriesClassifierPreset import TimeSeriesClassifierPreset
from fedot_ind.core.ensemble.static.RankEnsembler import RankEnsemble


class EnsembleGenerator(Enum):
    Rank_Ensemble = RankEnsemble


class TaskGenerator(Enum):

    ts_classification = dict(fedot_preset=TimeSeriesClassifierPreset,
                             nn=TimeSeriesClassifierNN,
                             default=TimeSeriesClassifier)

    image_classification = (TimeSeriesImageClassifier,)
    anomaly_detection = (TimeSeriesClassifier,)
    object_detection = (TimeSeriesClassifier,)



from torchvision.datasets import CIFAR10, CIFAR100, MNIST, ImageNet
from torchvision.transforms import Compose, Normalize, ToTensor, Resize


DATASETS_PARAMETERS = {
    "CIFAR100": {
        "getter": CIFAR100,
        "num_classes": 100,
        "train_ds_params": {
            "train": True,
            "download": True,
            "transform": Compose(
                [
                    ToTensor(),
                    Normalize((0.5074, 0.4867, 0.4411), (0.2011, 0.1987, 0.2025)),
                ]
            ),
        },
        "val_ds_params": {
            "train": False,
            "download": False,
            "transform": Compose(
                [
                    Resize((32, 32)),
                    ToTensor(),
                    Normalize((0.5074, 0.4867, 0.4411), (0.2011, 0.1987, 0.2025)),
                ]
            ),
        },
    },
    "CIFAR10": {
        "getter": CIFAR10,
        "num_classes": 10,
        "train_ds_params": {
            "train": True,
            "download": True,
            "transform": Compose(
                [
                    ToTensor(),
                    Normalize((0.5074, 0.4867, 0.4411), (0.2011, 0.1987, 0.2025)),
                ]
            ),
        },
        "val_ds_params": {
            "train": False,
            "download": False,
            "transform": Compose(
                [
                    Resize((32, 32)),
                    ToTensor(),
                    Normalize((0.5074, 0.4867, 0.4411), (0.2011, 0.1987, 0.2025)),
                ]
            ),
        },
    },
    "MNIST": {
        "getter": MNIST,
        "num_classes": 10,
        "train_ds_params": {
            "train": True,
            "download": True,
            "transform": Compose(
                [
                    ToTensor(),
                    Normalize((0.1307,), (0.3081,)),
                ]
            ),
        },
        "val_ds_params": {
            "train": False,
            "download": False,
            "transform": Compose(
                [
                    ToTensor(),
                    Normalize((0.1307,), (0.3081,)),
                ]
            ),
        },
    },
    "ImageNet": {
        "getter": ImageNet,
        "num_classes": 1000,
        "train_ds_params": {
            "split": "train",
            "transform": Compose(
                [
                    Resize((256, 265)),
                    ToTensor(),
                    Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
                ]
            ),
        },
        "val_ds_params": {
            "split": "val",
            "transform": Compose(
                [
                    Resize((256, 265)),
                    ToTensor(),
                    Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
                ]
            ),
        },
    },
}



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from fedot_ind.core.architecture.postprocessing.Analyzer import PerformanceAnalyzer
from fedot_ind.core.models.spectral.SSAExtractor import SSAExtractor
from fedot_ind.core.models.statistical.StatsExtractor import StatsExtractor
from fedot_ind.core.operation.optimization.FeatureSpace import VarianceSelector
from fedot_ind.core.architecture.experiment.TimeSeriesClassifier import TimeSeriesClassifier
from fedot_ind.core.architecture.preprocessing.DatasetLoader import DataLoader
from fedot_ind.core.operation.transformation.basis.legendre import LegenderBasis


def get_results(model, features, target):
    labels = model.predict(test_features=features)
    probs = model.predict_proba(test_features=features)
    metrics = PerformanceAnalyzer().calculate_metrics(target=target,
                                                      predicted_labels=labels['label'],
                                                      predicted_probs=probs['class_probability'])

    return metrics


def get_new_basis(train_data: tuple):
    data = []
    for i in range(0, train_data[0].shape[0], 1):
        x = train_data[0].iloc[i:i + 1, :].values.flatten()
        basis = LegenderBasis(data_range=len(x), n_components=10)
        basis.fit(x)
        basis.evaluate_derivative(order=1)
        # polynom_number = basis.analytical_form()
        # basis.show(visualisation_type='basis representation', basis_function=polynom_number)
        # basis.show(visualisation_type='derivative', basis_function=polynom_number)
        data.append(basis.basis[5])
        # fbasis = FourierDecomposition(data_range=len(x), n_components=10)
        # fbasis.decompose(x)

    update_train = pd.DataFrame(data)
    return update_train

dataset_name = 'Lightning7'
dataset = np.load(r'D:\РАБОТЫ РЕПОЗИТОРИИ\Репозитории\Industiral\IndustrialTS\data\UCI_HAR.npz')

test_data = dataset['data']

test_labels = dataset['labels']
train, test = DataLoader(dataset_name).load_data()
train_target = train[1]
test_target = test[1]
if __name__ == '__main__':
    df = pd.read_csv(r'C:\Users\user\Downloads\Telegram Desktop\12.11.22.csv')
    df.rename({'Base_metric': 'Metric_values_before_ensemble', 'Best_ensemble_metric': 'Metric_values_after_ensemble'},
              inplace=True,
              axis=1)
    df[df['Ensemble_models'] != '0'].boxplot(column=['Metric_values_before_ensemble', 'Metric_values_after_ensemble'])
    plt.show()
    update_train = get_new_basis(train_data=train)
    update_test = get_new_basis(train_data=test)

    spectral_model_basic = SSAExtractor(window_mode=True,
                                        window_sizes={dataset_name: [30]},
                                        spectral_hyperparams={'combine_eigenvectors': False,
                                                           'correlation_level': 0.8})
    train_feats_spectral_basic = spectral_model_basic.get_features(train[0], dataset_name)
    test_feats_spectral_basic = spectral_model_basic.get_features(test[0], dataset_name)
    train_eigenvectors_spectral_basic = spectral_model_basic.eigenvectors_list_train

    quantile_model_basic = StatsExtractor(window_mode=True)
    train_feats_baseline = quantile_model_basic.get_features(update_train, dataset_name)
    test_feats_baseline = quantile_model_basic.get_features(update_test, dataset_name)

    model_features = {
        'spectral_basic': train_feats_spectral_basic,
        'quantile': train_feats_baseline
    }

    strategy = VarianceSelector(models=model_features)

    best_model = strategy.get_best_model()
    model_data = model_features[best_model]
    projected_data = strategy.transform(model_data=model_data, principal_components=best_model)
    set_of_features = strategy.select_discriminative_features(model_data=model_data, projected_data=projected_data)
    for key, val in set_of_features.items():
        model_data[val].plot()
        plt.show()
    clf = TimeSeriesClassifier(model_hyperparams={
        'problem': 'classification',
        'seed': 42,
        'timeout': 2,
        'max_depth': 4,
        'max_arity': 2,
        'cv_folds': 2,
        'logging_level': 20,
        'n_jobs': 2
    })
    quantile_model_basic = StatsExtractor(window_mode=True)
    train_feats_baseline = quantile_model_basic.get_features(update_train, dataset_name)
    test_feats_baseline = quantile_model_basic.get_features(update_test, dataset_name)
    IndustrialModelBaseline = clf.fit(train_features=train_feats_baseline, train_target=train_target)[0]
    baseline_results = get_results(clf, test_feats_baseline, test_target)
    predict_baseline = clf.predict(test_feats_baseline)['label']



import gc
import sys
from typing import Optional

import numpy as np
import pandas as pd
from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.operation_parameters import OperationParameters
from gtda.time_series import takens_embedding_optimal_parameters
from scipy import stats
from tqdm import tqdm

from fedot_ind.core.models.BaseExtractor import BaseExtractor
from fedot_ind.core.operation.transformation.extraction.topological import AverageHoleLifetimeFeature, \
    AveragePersistenceLandscapeFeature, BettiNumbersSumFeature, HolesNumberFeature, MaxHoleLifeTimeFeature, \
    PersistenceDiagramsExtractor, PersistenceEntropyFeature, RadiusAtMaxBNFeature, RelevantHolesNumber, \
    SimultaneousAliveHolesFeature, SumHoleLifetimeFeature, TopologicalFeaturesExtractor

sys.setrecursionlimit(1000000000)

PERSISTENCE_DIAGRAM_FEATURES = {'HolesNumberFeature': HolesNumberFeature(),
                                'MaxHoleLifeTimeFeature': MaxHoleLifeTimeFeature(),
                                'RelevantHolesNumber': RelevantHolesNumber(),
                                'AverageHoleLifetimeFeature': AverageHoleLifetimeFeature(),
                                'SumHoleLifetimeFeature': SumHoleLifetimeFeature(),
                                'PersistenceEntropyFeature': PersistenceEntropyFeature(),
                                'SimultaneousAliveHolesFeature': SimultaneousAliveHolesFeature(),
                                'AveragePersistenceLandscapeFeature': AveragePersistenceLandscapeFeature(),
                                'BettiNumbersSumFeature': BettiNumbersSumFeature(),
                                'RadiusAtMaxBNFeature': RadiusAtMaxBNFeature()}


class TopologicalExtractor(BaseExtractor):
    """Class for extracting topological features from time series data.
    Args:
        params: parameters for operation

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.filtered_features = None
        self.feature_extractor = None

    def fit(self, input_data: InputData) -> OutputData:
        pass

    def generate_topological_features(self, ts_data: pd.DataFrame) -> pd.DataFrame:

        if self.feature_extractor is None:

            te_dimension, te_time_delay = self.get_embedding_params_from_batch(ts_data=ts_data)

            persistence_diagram_extractor = PersistenceDiagramsExtractor(takens_embedding_dim=te_dimension,
                                                                         takens_embedding_delay=te_time_delay,
                                                                         homology_dimensions=(0, 1),
                                                                         parallel=True)

            self.feature_extractor = TopologicalFeaturesExtractor(
                persistence_diagram_extractor=persistence_diagram_extractor,
                persistence_diagram_features=PERSISTENCE_DIAGRAM_FEATURES)

        ts_data_transformed = self.feature_extractor.fit_transform(ts_data)

        if self.filtered_features is None:
            self.filtered_features = ts_data_transformed.columns.tolist()
        gc.collect()
        return ts_data_transformed[self.filtered_features]

    def generate_features_from_ts(self, ts_data: pd.DataFrame, dataset_name: str = None):
        return self.generate_topological_features(ts_data=ts_data)

    def get_embedding_params_from_batch(self, ts_data: pd.DataFrame, method: str = 'mean') -> tuple:
        """Method for getting optimal Takens embedding parameters.
        Args:
            ts_data: dataframe with time series data
            method: method for getting optimal parameters
        Returns:
            Optimal Takens embedding parameters
        """
        methods = {'mode': self._mode,
                   'mean': np.mean,
                   'median': np.median}

        dim_list, delay_list = list(), list()

        for _ in tqdm(range(len(ts_data)),
                      initial=0,
                      desc='Time series processed: ',
                      unit='ts', colour='black'):
            ts_data = pd.DataFrame(ts_data)
            single_time_series = ts_data.sample(1, replace=False, axis=0).squeeze()
            delay, dim = takens_embedding_optimal_parameters(X=single_time_series,
                                                             max_time_delay=1,
                                                             max_dimension=5,
                                                             n_jobs=-1)
            delay_list.append(delay)
            dim_list.append(dim)

        dimension = int(methods[method](dim_list))
        delay = int(methods[method](delay_list))

        return dimension, delay

    @staticmethod
    def _mode(arr: list) -> int:
        return int(stats.mode(arr)[0][0])


import logging
import warnings

import numpy as np
import pandas as pd
from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from sklearn.metrics import accuracy_score, f1_score

warnings.simplefilter(action='ignore', category=FutureWarning)


class StaticBooster:
    CYCLES = 1

    def __init__(self, X_train, y_train, base_predict, timeout, threshold=0):

        self.X_train = X_train
        self.y_train = y_train
        self.base_predict = base_predict
        self.threshold = threshold
        self.timeout = round(timeout / 4)
        self.booster_features = {}
        self.check_table = pd.DataFrame()

        self.logger = logging.getLogger('StaticBooster')

    def run_boosting(self):
        self.logger.info('Started boosting')

        target_diff_1 = self.decompose_target(previous_predict=self.base_predict,
                                              previous_target=self.y_train.reshape(-1))
        prediction_1, model_1 = self.api_model(target_diff=target_diff_1, node='linear')

        target_diff_2 = self.decompose_target(previous_predict=prediction_1,
                                              previous_target=target_diff_1)
        prediction_2, model_2 = self.api_model(target_diff=target_diff_2, node='ridge')

        target_diff_3 = self.decompose_target(previous_predict=prediction_2,
                                              previous_target=target_diff_2)
        prediction_3, model_3 = self.api_model(target_diff=target_diff_3, node='lasso')

        final_prediction, model_ensemble = self.ensemble()

        try:
            self.check_table['target'] = self.proba_to_vector(self.y_train)
            self.check_table['final_predict'] = self.proba_to_vector(final_prediction)
            self.check_table['1_stage_predict'] = self.proba_to_vector(prediction_1)
            self.check_table['2_stage_predict'] = self.proba_to_vector(prediction_2)
            self.check_table['3_stage_predict'] = self.proba_to_vector(prediction_3)
            self.check_table['base_pred'] = self.proba_to_vector(self.base_predict)
        except Exception:
            self.logger.info('Problem with check table')

        model_list = [model_1, model_2, model_3]
        final_prediction_round = self.check_table['final_predict'].apply(func=np.round).values.reshape(-1)

        return final_prediction, model_list, model_ensemble

    def proba_to_vector(self, matrix):
        try:
            vector = np.array([x.argmax() + x[x.argmax()] for x in matrix])
            return vector
        except IndexError:
            return matrix

    def evaluate_results(self, target, prediction):

        if target.shape[0] > 2:
            average = 'weighted'
        else:
            average = 'binary'

        self.logger.info('Evaluation results')
        accuracy = accuracy_score(y_true=target,
                                  y_pred=prediction)
        f1 = f1_score(y_true=target,
                      y_pred=prediction,
                      average=average)

        return accuracy, f1

    def api_model(self, target_diff, node: str):
        self.logger.info(f'Starting cycle {self.CYCLES} of boosting')

        x_data = np.array(self.X_train)
        y_data = target_diff

        task = Task(TaskTypesEnum.regression)
        input_data = InputData(idx=np.arange(0, len(x_data)),
                               features=x_data,
                               target=y_data,
                               task=task,
                               data_type=DataTypesEnum.table)
        current_node = PrimaryNode(node)

        boosting_model = Pipeline(current_node)

        boosting_model.fit(input_data=input_data)
        prediction = boosting_model.predict(input_data=input_data)

        self.booster_features[self.CYCLES] = prediction.predict.reshape(-1)
        self.CYCLES += 1

        return prediction.predict.reshape(-1), boosting_model

    def decompose_target(self, previous_predict, previous_target):
        return previous_target - previous_predict

    def ensemble(self):
        self.logger.info('Starting to ensembling boosting results')

        x_data = pd.DataFrame.from_dict(self.booster_features, orient='index').T.values
        y_data = self.y_train

        task = Task(TaskTypesEnum.regression)

        input_data = InputData(idx=np.arange(0, len(x_data)),
                               features=x_data,
                               target=y_data,
                               task=task,
                               data_type=DataTypesEnum.table)

        xgboost = PrimaryNode('xgboost')
        ensemble_model = Pipeline(xgboost)

        ensemble_model.fit(input_data)
        ensemble_prediction = ensemble_model.predict(input_data)

        if np.unique(y_data).shape[0] > 2:
            ensemble_prediction = np.array([x.argmax() + x[x.argmax()] for x in ensemble_prediction.predict])
        else:
            ensemble_prediction = ensemble_prediction.predict.reshape(-1)

        return ensemble_prediction, ensemble_model

    def custom_round(self, num):
        thr = self.threshold
        if num - int(num) >= thr:
            return int(num) + 1
        return int(num)



from pathlib import Path
from typing import List

import setuptools

# The directory containing this file
HERE = Path(__file__).parent.resolve()

# The text of the README file
NAME = 'fedot_ind'
VERSION = '0.0.0.4'
AUTHOR = 'NSS Lab'
AUTHOR_EMAIL = 'itmo.nss.team@gmail.com'
SHORT_DESCRIPTION = 'Automated machine learning framework for time series analysis'
LONG_DESC_TYPE = 'text/x-rst'
README = Path(HERE, 'README_en.rst').read_text(encoding='utf-8')
EXCLUDED_PACKAGES = ['test*', 'examples', 'cases', 'benchmark']
URL = 'https://github.com/aimclub/Fedot.Industrial'
REQUIRES_PYTHON = '>=3.8'
LICENSE = 'BSD 3-Clause'
KEYWORDS = 'automated machine learning, time series analysis, anomaly detection, classification'


def _readlines(*names: str, **kwargs) -> List[str]:
    encoding = kwargs.get('encoding', 'utf-8')
    lines = Path(__file__).parent.joinpath(*names).read_text(encoding=encoding).splitlines()
    return list(map(str.strip, lines))


def _extract_requirements(file_name: str):
    return [line for line in _readlines(file_name) if line and not line.startswith('#')]


def _get_requirements(req_name: str):
    requirements = _extract_requirements(req_name)
    return requirements


setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=SHORT_DESCRIPTION,
    long_description=README,
    long_description_content_type=LONG_DESC_TYPE,
    url=URL,
    python_requires=REQUIRES_PYTHON,
    license=LICENSE,
    packages=setuptools.find_packages(exclude=EXCLUDED_PACKAGES),
    # package_data={'ucr_datasets': ['postprocessing/ucr_datasets.json']},
    include_package_data=True,
    install_requires=_get_requirements('requirements.txt'),
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords=KEYWORDS
)



# -*- coding: utf-8 -*-

from abc import ABC
from multiprocessing.dummy import Pool as ThreadPool

import numpy as np
import pandas as pd
from gtda.diagrams import BettiCurve, Filtering, PersistenceEntropy, PersistenceLandscape, Scaler
from gtda.homology import VietorisRipsPersistence
from gtda.time_series import TakensEmbedding


class PersistenceDiagramFeatureExtractor(ABC):
    """Abstract class persistence diagrams features extractor.

    """

    def extract_feature_(self, persistence_diagram):
        pass

    def fit_transform(self, x_pd):
        return np.array([self.extract_feature_(diagram) for diagram in x_pd])


class PersistenceDiagramsExtractor:
    """Class to extract persistence diagrams from time series.

    Args:
        takens_embedding_dim: Dimension of the Takens embedding.
        takens_embedding_delay: Delay of the Takens embedding.
        homology_dimensions: Homology dimensions to compute.
        filtering: Whether to filter the persistence diagrams.
        filtering_dimensions: Homology dimensions to filter.
        parallel: Whether to parallelize the computation.

    """

    def __init__(self, takens_embedding_dim: int,
                 takens_embedding_delay: int,
                 homology_dimensions: tuple,
                 filtering: bool = False,
                 filtering_dimensions: tuple = (1, 2),
                 parallel: bool = False):
        self.takens_embedding_dim_ = takens_embedding_dim
        self.takens_embedding_delay_ = takens_embedding_delay
        self.homology_dimensions_ = homology_dimensions
        self.filtering_ = filtering
        self.filtering_dimensions_ = filtering_dimensions
        self.parallel_ = parallel
        self.n_job = None

    def takens_embeddings_(self, data):
        te = TakensEmbedding(dimension=self.takens_embedding_dim_,
                             time_delay=self.takens_embedding_delay_)
        return te.fit_transform(data)

    def persistence_diagrams_(self, x_embeddings):
        if self.parallel_:
            pool = ThreadPool()
            x_transformed = pool.map(self.parallel_embed_, x_embeddings)
            pool.close()
            pool.join()
            return x_transformed
        else:
            x_transformed = list()
            for embedding in x_embeddings:
                x_transformed.append(self.parallel_embed_(embedding))
            return x_transformed

    def parallel_embed_(self, embedding):
        vr = VietorisRipsPersistence(metric='euclidean', homology_dimensions=self.homology_dimensions_,
                                     n_jobs=self.n_job)
        diagram_scaler = Scaler(n_jobs=self.n_job)
        persistence_diagrams = diagram_scaler.fit_transform(vr.fit_transform([embedding]))
        if self.filtering_:
            diagram_filter = Filtering(epsilon=0.1, homology_dimensions=self.filtering_dimensions_)
            persistence_diagrams = diagram_filter.fit_transform(persistence_diagrams)
        return persistence_diagrams[0]

    def fit_transform(self, x):
        x_embeddings = self.takens_embeddings_(x)
        x_persistence_diagrams = self.persistence_diagrams_(x_embeddings)
        return x_persistence_diagrams


class TopologicalFeaturesExtractor:
    def __init__(self, persistence_diagram_extractor, persistence_diagram_features):
        self.persistence_diagram_extractor_ = persistence_diagram_extractor
        self.persistence_diagram_features_ = persistence_diagram_features

    def fit_transform(self, x):

        x_pers_diag = self.persistence_diagram_extractor_.fit_transform(x)
        tmp = []
        column_list = []
        for feature_name, feature_model in self.persistence_diagram_features_.items():
            try:
                x_features = feature_model.fit_transform(x_pers_diag)
                tmp.append(x_features)
                for dim in range(len(x_features.shape)):
                    column_list.append('{}_{}'.format(feature_name, dim))
            except Exception:
                continue
        x_transformed = pd.DataFrame(data=np.hstack(tmp), columns=column_list)
        return x_transformed


class HolesNumberFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(HolesNumberFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        feature = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)
        for hole in persistence_diagram:
            if hole[1] - hole[0] > 0:
                feature[int(hole[2])] += 1.0
        return feature


class MaxHoleLifeTimeFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(MaxHoleLifeTimeFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        feature = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)
        for hole in persistence_diagram:
            lifetime = hole[1] - hole[0]
            if lifetime > feature[int(hole[2])]:
                feature[int(hole[2])] = lifetime
        return feature


class RelevantHolesNumber(PersistenceDiagramFeatureExtractor):
    def __init__(self, ratio=0.7):
        super(RelevantHolesNumber).__init__()
        self.ratio_ = ratio

    def extract_feature_(self, persistence_diagram):
        feature = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)
        max_lifetimes = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)

        for hole in persistence_diagram:
            lifetime = hole[1] - hole[0]
            if lifetime > max_lifetimes[int(hole[2])]:
                max_lifetimes[int(hole[2])] = lifetime

        for hole in persistence_diagram:
            index = int(hole[2])
            lifetime = hole[1] - hole[0]
            if np.equal(lifetime, self.ratio_ * max_lifetimes[index]):
                feature[index] += 1.0

        return feature


class AverageHoleLifetimeFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(AverageHoleLifetimeFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        feature = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)
        n_holes = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)

        for hole in persistence_diagram:
            lifetime = hole[1] - hole[0]
            index = int(hole[2])
            if lifetime > 0:
                feature[index] += lifetime
                n_holes[index] += 1

        for i in range(feature.shape[0]):
            feature[i] = feature[i] / n_holes[i] if n_holes[i] != 0 else 0.0

        return feature


class SumHoleLifetimeFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(SumHoleLifetimeFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        feature = np.zeros(int(np.max(persistence_diagram[:, 2])) + 1)
        for hole in persistence_diagram:
            feature[int(hole[2])] += hole[1] - hole[0]
        return feature


class PersistenceEntropyFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(PersistenceEntropyFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        persistence_entropy = PersistenceEntropy(n_jobs=-1)
        return persistence_entropy.fit_transform([persistence_diagram])[0]


class SimultaneousAliveHolesFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(SimultaneousAliveHolesFeature).__init__()

    @staticmethod
    def get_average_intersection_number_(segments):
        intersections = list()
        n_segments = segments.shape[0]

        for i in range(n_segments):
            count = 1
            start = segments[i, 0]
            end = segments[i, 1]

            for j in range(i + 1, n_segments):
                if start <= segments[j, 0] <= end:
                    count += 1
                else:
                    break
            intersections.append(count)

        return np.sum(intersections) / len(intersections)

    def get_average_simultaneous_holes_(self, holes):
        starts = holes[:, 0]
        ends = holes[:, 1]
        ind = np.lexsort((starts, ends))
        segments = np.array([[starts[i], ends[i]] for i in ind])
        return self.get_average_intersection_number_(segments)

    def extract_feature_(self, persistence_diagram):
        n_dims = int(np.max(persistence_diagram[:, 2])) + 1
        feature = np.zeros(n_dims)

        for dim in range(n_dims):
            holes = list()
            for hole in persistence_diagram:
                if hole[1] - hole[0] != 0.0 and int(hole[2]) == dim:
                    holes.append(hole)
            if len(holes) != 0:
                feature[dim] = self.get_average_simultaneous_holes_(np.array(holes))

        return feature


class AveragePersistenceLandscapeFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(AveragePersistenceLandscapeFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        # As practice shows, only 1st layer of 1st homology dimension plays role
        persistence_landscape = PersistenceLandscape(n_jobs=-1).fit_transform([persistence_diagram])[0, 1, 0, :]
        return np.array([np.sum(persistence_landscape) / persistence_landscape.shape[0]])


class BettiNumbersSumFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(BettiNumbersSumFeature).__init__()

    def extract_feature_(self, persistence_diagram):
        betti_curve = BettiCurve(n_jobs=-1).fit_transform([persistence_diagram])[0]
        return np.array([np.sum(betti_curve[i, :]) for i in range(int(np.max(persistence_diagram[:, 2])) + 1)])


class RadiusAtMaxBNFeature(PersistenceDiagramFeatureExtractor):
    def __init__(self):
        super(RadiusAtMaxBNFeature).__init__()

    def extract_feature_(self, persistence_diagram, n_bins=100):
        betti_curve = BettiCurve(n_jobs=-1, n_bins=n_bins).fit_transform([persistence_diagram])[0]
        max_dim = int(np.max(persistence_diagram[:, 2])) + 1
        max_bettis = np.array([np.max(betti_curve[i, :]) for i in range(max_dim)])
        return np.array(
            [np.where(betti_curve[i, :] == max_bettis[i])[0][0] / (n_bins * max_dim) for i in range(max_dim)])



from typing import Tuple, TypeVar, Optional

import numpy as np
import tensorly as tl
from fedot.core.operations.operation_parameters import OperationParameters
from pymonad.either import Either
from pymonad.list import ListMonad
from tensorly.decomposition import parafac

from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix
from fedot_ind.core.operation.transformation.regularization.spectrum import singular_value_hard_threshold, reconstruct_basis

class_type = TypeVar("T", bound="DataDrivenBasis")


class DataDrivenBasisImplementation(BasisDecompositionImplementation):
    """DataDriven basis
        Example:
            ts1 = np.random.rand(200)
            ts2 = np.random.rand(200)
            ts = [ts1, ts2]
            bss = DataDrivenBasisImplementation({'n_components': 3, 'window_size': 30})
            basis_multi = bss._transform(ts)
            basis_1d = bss._transform(ts1)
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.n_components = params.get('n_components')
        self.window_size = params.get('window_size')
        self.basis = None

    def _transform_one_sample(self, series: np.array):
        trajectory_transformer = HankelMatrix(time_series=series, window_size=self.window_size)
        data = trajectory_transformer.trajectory_matrix
        self.ts_length = trajectory_transformer.ts_length
        return self._get_basis(data)

    def _get_1d_basis(self, data):
        svd = lambda x: ListMonad(np.linalg.svd(x))
        threshold = lambda Monoid: ListMonad([Monoid[0],
                                              singular_value_hard_threshold(singular_values=Monoid[1],
                                                                            beta=data.shape[0] / data.shape[1],
                                                                            threshold=None),
                                              Monoid[2]]) if self.n_components is None else ListMonad([Monoid[0],
                                                                                                       Monoid[1][
                                                                                                       :self.n_components],
                                                                                                       Monoid[2]])
        data_driven_basis = lambda Monoid: ListMonad(reconstruct_basis(Monoid[0],
                                                                       Monoid[1],
                                                                       Monoid[2],
                                                                       ts_length=self.ts_length))

        basis = Either.insert(data).then(svd).then(threshold).then(data_driven_basis).value[0]

        return np.swapaxes(basis, 1, 0)

    def _get_multidim_basis(self, data):
        rank = round(data[0].shape[0] / 10)
        beta = data[0].shape[0] / data[0].shape[1]

        tensor_decomposition = lambda x: ListMonad(parafac(tl.tensor(x), rank=rank).factors)
        multi_threshold = lambda x: singular_value_hard_threshold(singular_values=x,
                                                                  beta=beta,
                                                                  threshold=None)

        threshold = lambda Monoid: ListMonad([Monoid[1],
                                              list(map(multi_threshold, Monoid[0])),
                                              Monoid[2].T]) if self.n_components is None else ListMonad([Monoid[1][
                                                                                                         :,
                                                                                                         :self.n_components],
                                                                                                         Monoid[0][
                                                                                                         :,
                                                                                                         :self.n_components],
                                                                                                         Monoid[2][
                                                                                                         :,
                                                                                                         :self.n_components].T])
        data_driven_basis = lambda Monoid: ListMonad(reconstruct_basis(Monoid[0],
                                                                       Monoid[1],
                                                                       Monoid[2],
                                                                       ts_length=self.ts_length))

        basis = np.array(
            Either.insert(data).then(tensor_decomposition).then(threshold).then(data_driven_basis).value[0])

        basis = basis.reshape(basis.shape[1], -1)

        return basis

    def evaluate_derivative(self:
                            class_type,
                            coefs: np.array,
                            order: int = 1) -> Tuple[class_type, np.array]:
        basis = type(self)(
            domain_range=self.domain_range,
            n_basis=self.n_basis - order,
        )
        derivative_coefs = np.array([np.polyder(x[::-1], order)[::-1] for x in coefs])

        return basis, derivative_coefs



from typing import Optional

from fedot.core.operations.evaluation.common_preprocessing import FedotPreprocessingStrategy
from fedot.core.operations.operation_parameters import OperationParameters

from fedot_ind.core.models.signal.RecurrenceExtractor import RecurrenceExtractor
from fedot_ind.core.models.signal.SignalExtractor import SignalExtractor
from fedot_ind.core.models.statistical.StatsExtractor import StatsExtractor
from fedot_ind.core.models.topological.TopologicalExtractor import TopologicalExtractor
from fedot_ind.core.operation.dummy.dummy_operation import DummyOperation
from fedot_ind.core.operation.transformation.basis.data_driven import DataDrivenBasisImplementation
from fedot_ind.core.operation.transformation.basis.fourier import FourierBasisImplementation
from fedot_ind.core.operation.transformation.basis.wavelet import WaveletBasisImplementation
from fedot_ind.core.repository.IndustrialOperationParameters import IndustrialOperationParameters


class IndustrialPreprocessingStrategy(FedotPreprocessingStrategy):
    """
    Args:
        operation_type: ``str`` of the operation defined in operation or data operation repositories

            .. details:: possible operations:

                - ``data_driven_basic``-> DataDrivenBasisImplementation,
                - ``topological_features``-> TopologicalExtractor,


        params: hyperparameters to fit the operation with

    """

    __operations_by_types = {
        'data_driven_basis': DataDrivenBasisImplementation,
        'wavelet_basis': WaveletBasisImplementation,
        'fourier_basis': FourierBasisImplementation,
        'topological_extractor': TopologicalExtractor,
        'quantile_extractor': StatsExtractor,
        'signal_extractor': SignalExtractor,
        'recurrence_extractor': RecurrenceExtractor,

        'cat_features': DummyOperation
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        params = IndustrialOperationParameters().from_params(operation_type, params) if params \
            else IndustrialOperationParameters().from_operation_type(operation_type)
        super().__init__(operation_type, params)

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain custom preprocessing strategy for {operation_type}')



from multiprocessing import Pool
from typing import Optional

from fedot.core.operations.operation_parameters import OperationParameters
from tqdm import tqdm

from fedot_ind.core.metrics.metrics_implementation import *
from fedot_ind.core.models.signal.WindowedFeaturesExtractor import WindowedFeatureExtractor
from fedot_ind.core.operation.transformation.DataTransformer import TSTransformer
from fedot_ind.core.operation.transformation.extraction.sequences import ReccurenceFeaturesExtractor


class RecurrenceExtractor(WindowedFeatureExtractor):
    """Class responsible for wavelet feature generator experiment.
    Args:
        window_mode: boolean flag - if True, window mode is used. Defaults to False.
        use_cache: boolean flag - if True, cache is used. Defaults to False.
    Attributes:
        transformer: TSTransformer object.
        self.extractor: ReccurenceExtractor object.
        train_feats: train features.
        test_feats: test features.
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.image_mode = False

        self.window_mode = params.get('window_mode')
        self.min_signal_ratio = params.get('min_signal_ratio')
        self.max_signal_ratio = params.get('max_signal_ratio')
        self.rec_metric = params.get('rec_metric')

        self.transformer = TSTransformer
        self.extractor = ReccurenceFeaturesExtractor
        self.train_feats = None
        self.test_feats = None

    def _ts_chunk_function(self, ts):
        specter = self.transformer(time_series=ts, min_signal_ratio=self.min_signal_ratio,
                                   max_signal_ratio=self.max_signal_ratio, rec_metric=self.rec_metric)
        feature_df = specter.ts_to_recurrence_matrix()
        if not self.image_mode:
            feature_df = pd.Series(self.extractor(recurrence_matrix=feature_df).recurrence_quantification_analysis())
        return feature_df

    def generate_vector_from_ts(self, ts_frame: pd.DataFrame) -> pd.DataFrame:
        """Generate vector from time series.
        Args:
            ts_frame: time series frame
        Returns:
            Feature vector
        """
        ts_samples_count = ts_frame.shape[0]
        n_processes = self.n_processes

        with Pool(n_processes) as p:
            components_and_vectors = list(tqdm(p.imap(self._ts_chunk_function,
                                                      ts_frame),
                                               total=ts_samples_count,
                                               desc='Feature Generation. TS processed',
                                               unit=' ts',
                                               colour='black'
                                               )
                                          )
        if self.image_mode:
            components_and_vectors = np.asarray(components_and_vectors)
            components_and_vectors = components_and_vectors[:, np.newaxis, :, :]
        else:
            components_and_vectors = pd.concat(components_and_vectors, axis=1).T
        return components_and_vectors

    def generate_features_from_ts(self, ts_data: pd.DataFrame,
                                  window_length: int = None) -> pd.DataFrame:

        if self.train_feats is None:
            train_feats = self.generate_vector_from_ts(ts_data)
            self.train_feats = train_feats
            return self.train_feats
        else:
            test_feats = self.generate_vector_from_ts(ts_data)
            if self.image_mode:
                self.test_feats = test_feats
            else:
                self.test_feats = test_feats[self.train_feats.columns]
            return self.test_feats



import numpy as np
import pandas as pd
import copy


def weighted_inner_product(F_i, F_j, window_length, ts_length, subseq_length):
    # Calculate the weights
    first = list(np.arange(window_length) + 1)
    second = [ts_length] * (subseq_length - window_length - 1)
    third = list(np.arange(window_length) + 1)[::-1]
    w = np.array(first + second + third)
    return w.dot(F_i * F_j)


def calculate_matrix_norms(rank, TS_comps, window_length, ts_length, subseq_length):
    F_wnorms = np.array(
        [weighted_inner_product(TS_comps[:, i], TS_comps[:, i], window_length, ts_length, subseq_length) for i in
         range(rank)])
    F_wnorms = F_wnorms ** -0.5
    return F_wnorms


def calculate_corr_matrix(rank, TS_comps, F_wnorms, window_length, ts_length, subseq_length):
    Wcorr = np.identity(rank)
    components = [i for i in range(rank)]
    for i in components:
        for j in range(i + 1, rank):
            Wcorr[i, j] = abs(
                weighted_inner_product(TS_comps[:, i], TS_comps[:, j], window_length, ts_length, subseq_length) *
                F_wnorms[i] * F_wnorms[j])
            Wcorr[j, i] = Wcorr[i, j]
    return Wcorr, components


def combine_eigenvectors(TS_comps, rank, window_length, ts_length, subseq_length, correlation_level: float = 0.8):
    """Calculates the w-correlation matrix for the time series.

    Args:
        TS_comps (np.ndarray): The time series components.
        rank (int): The rank of the time series.
        correlation_level (float): threshold value of Pearson correlation, using for merging eigenvectors.
        subseq_length (int): The length of TS subseq.
        ts_length (int): The length of TS .
        window_length (int): The length of TS window.


    """
    combined_components = []

    # Calculated weighted norms
    F_wnorms = calculate_matrix_norms(rank, TS_comps, window_length, ts_length, subseq_length)

    # Calculate Wcorr.
    Wcorr, components = calculate_corr_matrix(rank, TS_comps, F_wnorms, window_length, ts_length, subseq_length)

    # Calculate Wcorr. and Select Correlated Eigenvectors.
    corr_dict = {i: [i for i, v in enumerate(Wcorr[:, i]) if v > correlation_level] for i in components}
    # copy of the dictionary for deleting keys
    filtred_dict = copy.deepcopy(corr_dict)

    # Select Correlated Eigenvectors.
    for list_of_corr_vectors in list(corr_dict.values()):
        final_component = None
        if len(list_of_corr_vectors) < 2:
            final_component = TS_comps[:, list_of_corr_vectors[0]]
            combined_components.append(final_component)
        else:
            for corr_vector in list_of_corr_vectors:
                if corr_vector in filtred_dict.keys():
                    if final_component is None:
                        final_component = np.sum(TS_comps[:, list_of_corr_vectors], axis=1)
                        combined_components.append(final_component)
                    del filtred_dict[corr_vector]

    combined_components = pd.DataFrame(combined_components).T

    return combined_components



from fedot_ind.core.models.detection.utils.get_time \
    import time_now
from fedot_ind.core.models.detection.abstract_objects.AbstractDataOperation import AbstractDataOperation

from fedot_ind.core.models.detection.utils.math_utils import NormalizeData, Cut_data

from fedot_ind.core.models.detection.abstract_objects.AnomalyZone import AnomalyZone
from fedot_ind.core.models.detection.abstract_objects.FileObject import FileObject

"""

input format:

    dict with "data" and "lables" fields

Output 
    the same dict but with additional list of window
    
"""
from typing import List, Type

class ThresholdZonesDetector(AbstractDataOperation):
    """
    This detector cut ts up to demanded percentage of 0 in it and 
    make detection of anomaly zones that could be used in future

    Args:
        AbstractDetector (_type_): _description_
    """
    def __init__(self, zero_percentage = 0.5,
            ignore_first: bool = False,
            ignore_last: bool = False,
            detector_name: str = "Threshold Zones Detector"):
        self.ignore_last: bool = ignore_last
        self.ignore_first: bool = ignore_first
        self.target_percentage: float = zero_percentage
        self.name: str = detector_name
        self.print_logs: bool = True

    def set_settings(self, print_logs: bool = True):
        self.print_logs = print_logs
        self._print_logs("Settings was set.")
        self._print_logs(f"Print logs = {print_logs}")

    def load_data(self, file_object: FileObject) -> None:
        self._print_logs("Data read!")
        self.file_object: FileObject = file_object
        #self.input_dict[DATA_BODY][ELECTED_DATA]
        #self.lables = self.input_dict[DATA_BODY][RAW_LABLES]

    def run_operation(self) -> None:
        self._print_logs("Start detection...")
        self._vector_analysis()
        self._print_logs("Detection finished!")

    def return_new_data(self) -> FileObject:
        return self.file_object

    def get_percentage_of_zero(self, time_series: List[float]) -> float:
        """

        Args:
            time_series (list): _description_

        Returns:
            float: percentage of 0 in list
        """
        return time_series.count(0)/len(time_series)

    def _vector_analysis(self) -> None:   
        self.output_predicts = []
        self.output_quantile_predicts = []
        self.output_quantile_predicts_for_show = []
        
        self.target_error = 0.05
        self.low_thresh = self.target_percentage - self.target_error
        self.high_thresh = self.target_percentage + self.target_error
        self._print_logs(f"Predict on {self.file_object.filename}...")

        current_thresh = 0.3
        step = 0.005
        self.file_object.anomalies_list = []
        # -------------------------------------
        temp_data_exp = self.file_object.test_data

        counter = 0
        self._print_logs(f"Predict on {self.file_object.filename}, attempt to transform data to {self.target_percentage} of zero...")

        while True:
            counter+=1
            temp_reshaped_data = Cut_data(temp_data_exp, current_thresh)
            if self.low_thresh <= self.get_percentage_of_zero(temp_reshaped_data)<=self.high_thresh or counter >= 200:
                temp_data_exp = temp_reshaped_data
                self.file_object.threshold = current_thresh
                self._print_logs(f"Predict on {self.file_object.filename}, succes! Current threshold: {current_thresh}")
                break
            else:
                if self.get_percentage_of_zero(temp_reshaped_data) > self.target_percentage:
                    current_thresh-=step
                else:
                    current_thresh+=step
        self.file_object.additional_transformed_average_absolute_deviation = \
            temp_data_exp



        areas_list, len_list, distances_list = self._get_areas_of_data(temp_data_exp)
        ensambled_prediction: List[Type[AnomalyZone]] = []
        self._print_logs(f"Predict on {self.file_object.filename}, filtering...")

        if len(areas_list) and len(len_list) and len(distances_list):
            for area in areas_list:
                if area[1] - area[0] >= 100:
                    ensambled_prediction.append(AnomalyZone(area[0], area[1]))

        self.file_object.anomalies_list = ensambled_prediction

        for j in range(self.file_object.get_len_of_dataset()):
            if self.file_object.additional_transformed_average_absolute_deviation[j] != 0:
                self.file_object.additional_transformed_average_absolute_deviation[j] = \
                    self.file_object.additional_transformed_average_absolute_deviation[j] - current_thresh
        self.file_object.additional_transformed_average_absolute_deviation = \
            NormalizeData(self.file_object.additional_transformed_average_absolute_deviation)

        self.file_object.copy_parts_of_datasets_in_anomaly_zones()



    def _get_areas_of_data(self, data: list):
        """

        Args:
            data (list): _description_
        """
        # returns three lists
        # 1 - we are in, 0 - we are out, 2 - start
        state_of_scaning: int = 2
        counter_for_areas = 0
        areas_list = []
        len_list = []
        areas_sizes_list = []
        max_list = []
        max_dist = 0
        area = 0
        start = 0
        end = 0

        
        for i in range(len(data)):
            if state_of_scaning == 2:
                if data[i] != 0:
                    state_of_scaning = 1
                    max_dist = max(max_dist, data[i])
                    start = i
                else:
                    state_of_scaning = 0
                    start = i
                continue

            if state_of_scaning == 1:
                if data[i] == 0: 
                    state_of_scaning = 0
                    end = i
                    areas_list.append([start, end])
                    max_list.append(max_dist)
                    len_list.append(counter_for_areas)
                    areas_sizes_list.append(area)
                    counter_for_areas = 0
                    max_dist = 0
                    start = i
                    area = 0
                else:
                    counter_for_areas += 1

                    area += data[i]
                    max_dist = max(max_dist, data[i])  
                continue
            if state_of_scaning == 0:
                if data[i] != 0:  
                    start = i
                    state_of_scaning = 1
                    counter_for_areas = 1
                    area = 0
                    max_dist = max(max_dist, data[i])
                continue
        if state_of_scaning == 1:
            areas_list.append([start, len(data)])
            max_list.append(max_dist)
            len_list.append(counter_for_areas)
            areas_sizes_list.append(area)
        if self.ignore_first:
            del areas_list[0]
            del max_list[0]
            del len_list[0]
            del areas_sizes_list[0]
        if self.ignore_last:
            del areas_list[-1]
            del max_list[-1]
            del len_list[-1]
            del areas_sizes_list[-1]

        return areas_list, len_list, max_list


    def _print_logs(self, log_message: str) -> None:
        if self.print_logs:
            print(f"---[{time_now()}] {self.name}: {log_message}")


from typing import Optional

import pandas as pd

from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters

from fedot_ind.core.models.BaseExtractor import BaseExtractor
from fedot_ind.core.operation.transformation.extraction.statistical import StatFeaturesExtractor


class StatsExtractor(BaseExtractor):
    """Class responsible for quantile feature generator experiment.
    Args:
        window_mode: Flag for window mode. Defaults to False.
        use_cache: Flag for cache usage. Defaults to False.
    Attributes:
        use_cache (bool): Flag for cache usage.
        aggregator (StatFeaturesExtractor): StatFeaturesExtractor object.
        vis_flag (bool): Flag for visualization.
        train_feats (pd.DataFrame): Train features.
        test_feats (pd.DataFrame): Test features.
    """
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.aggregator = StatFeaturesExtractor()
        self.window_mode = params.get('window_mode')
        self.window_size = params.get('window_size')
        self.vis_flag = False
        self.train_feats = None
        self.test_feats = None
        self.n_components = None

    def fit(self, input_data: InputData):
        pass

    def extract_stats_features(self, ts):
        if self.window_mode:
            aggregator = self.aggregator.create_baseline_features
            list_of_stat_features_on_interval = self.apply_window_for_stat_feature(ts_data=ts,
                                                                                   feature_generator=aggregator,
                                                                                   window_size=self.window_size)
            aggregation_df = pd.concat(list_of_stat_features_on_interval, axis=1)
        else:
            aggregation_df = self.aggregator.create_baseline_features(ts)
        return aggregation_df

    def generate_features_from_ts(self,
                                  ts_frame: pd.DataFrame,
                                  window_length: int = None) -> pd.DataFrame:

        ts = pd.DataFrame(ts_frame, dtype=float)
        ts = ts.fillna(method='ffill')

        if ts.shape[0] == 1 or ts.shape[1] == 1:
            mode = 'SingleTS'
        else:
            mode = 'MultiTS'
        try:
            if mode == 'MultiTS':
                aggregation_df = self.__get_feature_matrix(ts)
            else:
                aggregation_df = self.extract_stats_features(ts)
        except Exception:
            aggregation_df = self.__component_extraction(ts)

        return aggregation_df

    def __component_extraction(self, ts):
        ts_components = [pd.DataFrame(x) for x in ts.T.values.tolist()]
        tmp_list = []
        for index, component in enumerate(ts_components):
            aggregation_df = self.extract_stats_features(component)
            col_name = [f'{x} for component {index}' for x in aggregation_df.columns]
            aggregation_df.columns = col_name
            tmp_list.append(aggregation_df)
        aggregation_df = pd.concat(tmp_list, axis=1)
        return aggregation_df

    def __get_feature_matrix(self, ts):
        ts_components = [pd.DataFrame(x) for x in ts.values.tolist()]
        if ts_components[0].shape[0] != 1:
            ts_components = [x.T for x in ts_components]
        tmp_list = []
        for index, component in enumerate(ts_components):
            aggregation_df = self.extract_stats_features(component)
            tmp_list.append(aggregation_df)
        aggregation_df = pd.concat(tmp_list, axis=0)
        return aggregation_df


from typing import Tuple

import numpy as np
import torch
from torch.utils.data import Dataset


class CustomClassificationDataset(Dataset):
    """Class for custom classification datasets.

    Args:
        images: Numpy matrix of images.
        targets: Numpy vector of targets.
    """

    def __init__(
            self,
            images: np.ndarray,
            targets: np.ndarray,
    ) -> None:
        self.images = torch.from_numpy(images).float()
        self.targets = torch.from_numpy(targets).type(torch.long)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, int]:
        """Returns a sample from a dataset.

        Args:
            idx: Index of sample.

        Returns:
            A tuple ``(image, target)``, where image is image tensor,
                and target is integer.
        """
        return self.images[idx], self.targets[idx]

    def __len__(self) -> int:
        """Return length of dataset"""
        return self.targets.size()[0]